! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: esas_mod
!
! Authors: Brendan Perry - NASA/GSFC F77-f95 conversion and SAS compliance
!          Steve Snowden - NASA/GSFC (original algorithms and coding)
!          Kip Kuntz     - NASA/GSFC/JHU (original algorithms)
!          
! Comments: subroutine library for use by espfilt and new ESAS tasks (SAS-19)
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
! *** BP rewrite of Snowden esas "subs.f" routines ***
!     for SAS/F90/gfortran compliance
!
module esas_mod

  use cal
  use dal
  use param 
  use errorHandling
  use dataFilter
  use gofio_mod
  use gofutils_mod
  use gofmatrix_mod
  use gofcoords_mod
  use, intrinsic :: ieee_arithmetic ! use to weed out nulls

  implicit none

contains
  !
  ! was aug_get_exp in ESAS
  !
  subroutine augment_mos (qdp_F,QPB_S,ratein,hardin,revlin,c_dst,c_tim, &
       doplot,aug_spec,aug_expo,mosnum,ccdN,chans,obsid,expid,firsttime,&
       lunqdp)
    !
    ! called by ESAS routine(s):
    !  mosback (old mos_back)
    !
    ! Inputs:
    !   qdp_F: QDP plot file name (mos1S001-aug.qdp)
    !   QPB_S: quiescent particle background CCF set already opened
    !   ratein: R*4 rate from mosback old c_val(1)
    !   hardin: R*4 hard from mosback old c_val(2)
    !   revlin: I*4 revl from mosback old c_val(3)
    !   c_dst(3) (/0.01,1.,0.1/) hardcoded in mosback R*4
    !   c_tim    1E06 hardcoded in mosback            R*4
    !   doplot   usually 'withqdp' logical  L*1
    !   aug_spec(chans) OUT augmented spectra R*4
    !   aug_expo        OUT augmented exposure R*4
    !   mosnum          1 or 2 I*4 
    !   ccdN            2-7    I*4
    !   chans           2400 for mosback, variable I*4
    !   obsid           1234567890 C*10
    !   expid           1S002      C*5
    !   firsttime       first time thru here? L*1 for opening QDP file
    !   lunqdp          Logical unit number for output QDP file I*4
    !
    type(DataSetT), intent(inout) :: QPB_S
    type(TableT)   :: QPB_T
    type(ColumnT)  :: obsid_C, expid_C, object_C, date_C, back_C, &
         revl_C, filter_C, expo_C, rate_C, submode_C, chip_C, &
         urate_C, hard_C, uhard_C, spec_C, ttime_C, key_C
    !
    character(len=*), intent(in) :: qdp_F, expid, obsid
    character(len=24) :: title
    character(len=20), allocatable, dimension(:) :: obid, objt
    character(len=20) :: extname
    character(len=15) :: ecode
    character(len=10), dimension(:), allocatable :: obsidn
    character(len=6), dimension(:), allocatable :: filt
    character(len=1) :: cccd, mosno
    integer, intent(in) :: chans, ccdN, mosnum, revlin
    integer, intent(inout) :: lunqdp
    integer(kind=int16), dimension(:), pointer :: chip_IP
    integer(kind=int32), dimension(:,:), pointer :: spec_JP
    integer(kind=int32), dimension(:), pointer :: revl_JP
    integer :: i, ii, j, nn, ikeep, iodd, itd, modd, nchan, chipRows(7), &
         ncols, nobs, n, qpbRows, newRows, status = 0
    integer, allocatable, dimension(:) :: revl, ord
    logical, intent(in) :: doplot
    logical, intent(inout) :: firsttime
    logical :: warndie = .TRUE.
    real(kind=single), intent(in) :: c_tim, c_dst(3), ratein, hardin
    real(kind=single), intent(inout) :: aug_expo
    real(kind=single) :: ave_rate, cum_exp, d1, d2, hardmax, hardmin, &
         muhard, murate, ratemax, ratemin, td, temphard, temp, tr, exx
    !
    ! BHP: old was 2400 channel spectra and 15000 QPB rows
    !      now, chans spectra rows are allocatable
    real(kind=single) :: aug_spec(chans) ! 2400
    real(kind=single), allocatable, dimension(:) :: ddist, dist,&
         energ, expo, hard, hdist, rate, rdist, tdist, time, &
         uhard, urate
    ! old 7500: date,ddist,dist,expo,hard,hdist,rate,rdist,tdist,
    !           time,uhard,urate
    ! old 2400: energ
    !    BHP: do all pointer math now...
    real(kind=double), dimension(:), pointer :: expo_DP, rate_DP, &
         urate_DP, hard_DP, uhard_DP, date_DP
    !
    ! qdpfile was augfile
    ! create title for qdp plot
    !
    write(cccd,'(I1)') ccdN ! current CCD number in character*1
    write(mosno,'(I1)') mosnum
    title = '"' // adjustl(trim(obsid)) // ' -- MOS' // &
         mosno // " " // adjustl(trim(expid)) // '"'
    !
    !  SS: If selected, make a plot, first set up the plot
    !
    if(doplot) then
       if (firsttime) then
          write(errstr,*)"Writing out QDP plot file: ",trim(qdp_F)
          call message(AppMsg,Noisy)
          !
          ! open the qdp file for writing
          !
          call open_ascii_file(qdp_F,lunqdp)
          write(lunqdp,*) ' viewport 0.2 0.2 0.9 0.8'
          write(lunqdp,*) ' la t ', title
          write(lunqdp,*) ' read serr 1'
          write(lunqdp,*) ' read serr 2'
          write(lunqdp,*) ' read serr 3'
          write(lunqdp,*) ' read serr 4'
          write(lunqdp,*) ' read serr 5'
          write(lunqdp,*) ' read serr 6'
          write(lunqdp,*) ' read serr 7'
          write(lunqdp,*) ' read serr 8'
          write(lunqdp,*) ' read serr 9'
          write(lunqdp,*) ' read serr 10'
          write(lunqdp,*) ' read serr 11'
          write(lunqdp,*) ' read serr 12'
          write(lunqdp,*) ' read serr 13'
          write(lunqdp,*) ' read serr 14'
          write(lunqdp,*) ' read serr 15'
          write(lunqdp,*) ' read serr 16'
          write(lunqdp,*) ' read serr 17'
          write(lunqdp,*) ' read serr 18'
          write(lunqdp,*) ' read serr 19'
          write(lunqdp,*) ' font roman'
          write(lunqdp,*) ' cs 1.5'
          write(lunqdp,*) ' la f'
          write(lunqdp,*) ' col 2 on 2'
          write(lunqdp,*) ' col 1 on 3'
          write(lunqdp,*) ' col 3 on 4'
          write(lunqdp,*) ' col 2 on 5'
          write(lunqdp,*) ' col 1 on 6'
          write(lunqdp,*) ' col 3 on 7'
          write(lunqdp,*) ' col 2 on 8'
          write(lunqdp,*) ' col 1 on 9'
          write(lunqdp,*) ' col 3 on 10'
          write(lunqdp,*) ' col 2 on 11'
          write(lunqdp,*) ' col 1 on 12'
          write(lunqdp,*) ' col 3 on 13'
          write(lunqdp,*) ' col 2 on 14'
          write(lunqdp,*) ' col 1 on 15'
          write(lunqdp,*) ' col 3 on 16'
          write(lunqdp,*) ' col 2 on 17'
          write(lunqdp,*) ' col 1 on 18'
          write(lunqdp,*) ' col 3 on 19'
          write(lunqdp,*) ' mark 1 on 2'
          write(lunqdp,*) ' mark 1 on 3'
          write(lunqdp,*) ' mark 1 on 4'
          write(lunqdp,*) ' mark 1 on 5'
          write(lunqdp,*) ' mark 1 on 6'
          write(lunqdp,*) ' mark 1 on 7'
          write(lunqdp,*) ' mark 1 on 8'
          write(lunqdp,*) ' mark 1 on 9'
          write(lunqdp,*) ' mark 1 on 10'
          write(lunqdp,*) ' mark 1 on 11'
          write(lunqdp,*) ' mark 1 on 12'
          write(lunqdp,*) ' mark 1 on 13'
          write(lunqdp,*) ' mark 1 on 14'
          write(lunqdp,*) ' mark 1 on 15'
          write(lunqdp,*) ' mark 1 on 16'
          write(lunqdp,*) ' mark 1 on 17'
          write(lunqdp,*) ' mark 1 on 18'
          write(lunqdp,*) ' mark 1 on 19'
          write(lunqdp,*) ' line off 2'
          write(lunqdp,*) ' line off 3'
          write(lunqdp,*) ' line off 4'
          write(lunqdp,*) ' line off 5'
          write(lunqdp,*) ' line off 6'
          write(lunqdp,*) ' line off 7'
          write(lunqdp,*) ' line off 8'
          write(lunqdp,*) ' line off 9'
          write(lunqdp,*) ' line off 10'
          write(lunqdp,*) ' line off 11'
          write(lunqdp,*) ' line off 12'
          write(lunqdp,*) ' line off 13'
          write(lunqdp,*) ' line off 14'
          write(lunqdp,*) ' line off 15'
          write(lunqdp,*) ' line off 16'
          write(lunqdp,*) ' line off 17'
          write(lunqdp,*) ' line off 18'
          write(lunqdp,*) ' line off 19'
          write(lunqdp,*) ' win 1'
          write(lunqdp,*) ' yplot 1 2 3 4'
          write(lunqdp,*) ' loc 0.05 0.48 0.39 1.0'
          write(lunqdp,*) ' la 11 p 0.06 9.0 just left "CCD #2"'
          write(lunqdp,*) ' r x1 0.0 0.1'
          write(lunqdp,*) ' r y1 0.0 10.0'
          write(lunqdp,*) ' la y Hardness'
          write(lunqdp,*) ' win 2'
          write(lunqdp,*) ' yplot 1 5 6 7'
          write(lunqdp,*) ' loc 0.35 0.48 0.69 1.0'
          write(lunqdp,*) ' la 12 p 0.06 9.0 just left "CCD #3"'
          write(lunqdp,*) ' r x2 0.0 0.1'
          write(lunqdp,*) ' r y2 0.0 10.0'
          write(lunqdp,*) ' win 3'
          write(lunqdp,*) ' yplot 1 8 9 10'
          write(lunqdp,*) ' loc 0.65 0.48 0.99 1.0'
          write(lunqdp,*) ' la 13 p 0.06 9.0 just left "CCD #4"'
          write(lunqdp,*) ' r x3 0.0 0.1'
          write(lunqdp,*) ' r y3 0.0 10.0'
          write(lunqdp,*) ' win 4'
          write(lunqdp,*) ' yplot 1 11 12 13'
          write(lunqdp,*) ' loc 0.05 0.05 0.39 0.57'
          write(lunqdp,*) ' la 14 p 0.06 9.0 just left "CCD #5"'
          write(lunqdp,*) ' r x4 0.0 0.1'
          write(lunqdp,*) ' r y4 0.0 10.0'
          write(lunqdp,*) ' la y Hardness'
          write(lunqdp,*) ' la x Rate'
          write(lunqdp,*) ' win 5'
          write(lunqdp,*) ' yplot 1 14 15 16'
          write(lunqdp,*) ' loc 0.35 0.05 0.69 0.57'
          write(lunqdp,*) ' la 15 p 0.06 9.0 just left "CCD #6"'
          write(lunqdp,*) ' r x5 0.0 0.1'
          write(lunqdp,*) ' r y5 0.0 10.0'
          write(lunqdp,*) ' la x Rate'
          write(lunqdp,*) ' win 6'
          write(lunqdp,*) ' yplot 1 17 18 19'
          write(lunqdp,*) ' loc 0.65 0.05 0.99 0.57'
          write(lunqdp,*) ' la 16 p 0.06 9.0 just left "CCD #7"'
          write(lunqdp,*) ' r x6 0.0 0.1'
          write(lunqdp,*) ' r y6 0.0 10.0'
          write(lunqdp,*) ' la x Rate'
       end if
    endif
    !
    extname = "QPB" 
    if (.not.hasBlock(QPB_S,trim(adjustl(extname)))) then ! should never happen!!!
       call error("noQPB", "CAL QPB file does not contain QPB extension")
    else
       QPB_T = table(QPB_S,trim(adjustl(extname))) ! read the whole extension
       qpbRows = numberOfRows(QPB_T)               ! get number of rows
       if (firsttime) then
          write(errstr,*)"CAL QPB file has ",qpbRows, " rows"
          call message(AppMsg,Always) ! change this ???!!!
       endif
    endif
    !
    !*** now read the colums in emos#_qpb.fits.gz file (BHP) ***
    !*** files are in packages/testprods/esas_data
    !*** Columns = obsid, name, object, date, revolut, filter, ontime,
    !*** expo, rate, rate_err, hard, hard_err, spectrum
    !
    write(errstr,*)"Reading relevant columns from QPB SPECTRUM extension"
    if (.not.hasColumn(QPB_T,"CHIP")) then
       call error("badColumn","Input file incorrect - no CHIP column")
    else
       chip_C = column(QPB_T,"CHIP",READ)
       chip_IP => int16Data(chip_C)
    endif
    if (.not.hasColumn(QPB_T,"DATE")) then
       call error("badColumn","Input file incorrect - no DATE column")
    else
       date_C = column(QPB_T,"DATE",READ)
       date_DP => real64Data(date_C)
    endif
    if (.not.hasColumn(QPB_T,"REVOLUT")) then
       call error("badColumn","Input file incorrect - no REVOLUT column")
    else
       revl_C = column(QPB_T,"REVOLUT",READ)
       revl_JP => int32Data(revl_C)
    endif
    if (.not.hasColumn(QPB_T,"OBSID")) then
       call error("badColumn","Input file incorrect - no OBSID column")
    else
       obsid_C = column(QPB_T,"OBSID",READ)
       allocate(obsidn(qpbRows))
       do i=1,qpbRows
          obsidn(i) = stringCell(obsid_C,i-1)  
       end do
    endif
    if (.not.hasColumn(QPB_T,"FILTER")) then
       call error("badColumn","Input file incorrect - no FILTER column")
    else
       filter_C = column(QPB_T,"FILTER",READ)
       allocate(filt(qpbRows))
       do i=1,qpbRows
          filt(i) = stringCell(filter_C,i-1)  ! fill chans array with keys
       end do
    endif
    if (.not.hasColumn(QPB_T,"EXPOSURE")) then
       call error("badColumn","Input file incorrect - no EXPOSURE column")
    else
       expo_C = column(QPB_T,"EXPOSURE",READ)
       expo_DP => Real64Data(expo_C)
    endif
    if (.not.hasColumn(QPB_T,"RATE")) then
       call error("badColumn","Input file incorrect - no RATE column")
    else
       rate_C = column(QPB_T,"RATE",READ)
       rate_DP => Real64Data(rate_C)
    endif
    if (.not.hasColumn(QPB_T,"RATEERR")) then
       call error("badColumn","Input file incorrect - no RATEERR column")
    else
       urate_C = column(QPB_T,"RATEERR",READ)
       urate_DP => Real64Data(urate_C)
    endif
    if (.not.hasColumn(QPB_T,"HARD")) then
       call error("badColumn","Input file incorrect - no HARD column")
    else
       hard_C = column(QPB_T,"HARD",READ)
       hard_DP => Real64Data(hard_C)
    endif
    if (.not.hasColumn(QPB_T,"HARDERR")) then
       call error("badColumn","Input file incorrect - no HARDERR column")
    else
       uhard_C = column(QPB_T,"HARDERR",READ)
       uhard_DP => Real64Data(uhard_C)
    endif
    if (.not.hasColumn(QPB_T,"SPECTRUM")) then
       call error("badColumn","Input file incorrect - no SPECTRUM column")
    else
       spec_C = column(QPB_T,"SPECTRUM",READ)
       spec_JP => int32VectorData(spec_C)
    endif
    nn = 0
    do i=0,qpbRows-1
       if (chip_IP(i) .eq. ccdN) nn = nn + 1
    enddo
    chipRows(ccdN) = nn
    write(errstr,1050)"CCD",ccdN," is ",nn," rows of CCF QPB file."
    call message(AppMsg,Always) ! change this ???!!!
1050 format(A,I1,A,I5,A)
    !
    !  SS: Do some cleaning, and selecting on CCD state
    !
    !  SS: First determine if the observation CCD is in an odd state
    !
    iodd = 0
    tr = 0.
    temphard = 0.
    if(mosnum .eq. 1) then
       if (ccdN .eq. 4) then
          tr = 100.0 * ratein ! OLDNEW: c_val(1)/ratein
          if(tr .lt. 5.2) then
             temphard = 0.96
          elseif(tr .lt. 5.2) then
             temphard = 0.423*tr - 1.24
          else
             temphard = 0.423*tr - 1.24
          endif
          if(hardin .lt. temphard) iodd = 1 ! OLDNEW: c_val(2)/hardin
       elseif (ccdN .eq. 5) then
          tr = 100.0 * ratein
          if(tr .lt. 7.0) then
             temphard = 1.54
          elseif (tr .lt. 9.0) then
             temphard = 0.53*tr - 2.17
          else
             temphard = 2.60
          endif
          if(hardin .lt. temphard) iodd = 1
       else
          if(hardin .lt. 1.5) iodd = 1
       endif
    else
       if (ccdN .eq. 2) then
          if(hardin .lt. 1.1) iodd = 1
       elseif (ccdN .eq. 5) then
          tr = 100.0 * ratein
          if(tr .lt. 10.0) then
             temphard = 0.15*tr
          elseif (tr .lt. 16.67) then
             temphard = 0.15*tr
          else
             temphard = 2.50
          endif
          if(hardin .lt. temphard) iodd = 1
       else
          if(hardin .lt. 1.5) iodd = 1
       endif
    endif
    !print*,'tr = ',tr,' temphard = ',temphard,' hardin = ',hardin,' iodd = ',iodd
    !
    !  SS: Set the normal filtering limits
    !
    d1 = -9.0
    d2 = 0.0
    hardmin = 0.2
    hardmax = 15.0 ! hardcoded or parameters??? BHP
    ratemin = 0.01
    ratemax = 0.1
    !
    ! SS: Sort through the archived observation data
    !
    nobs = 0
    exx = 0.
    do i=0,qpbRows-1 ! OLDNEW: nchan/qpbRows fix this???!!!
       if (chip_IP(i) .ne. ccdN) cycle
       if(  (expo_DP(i) .ge. 1000.0)  .and. &
            (rate_DP(i) .gt. ratemin) .and. &
            (rate_DP(i) .lt. ratemax) .and. &
            (hard_DP(i) .gt. hardmin) .and. &
            (hard_DP(i) .lt. hardmax) .and. &
            (filt(i+1)  .ne. 'CalClo')) then
          !
          !  Write out all the data for plotting if selected
          !
          if(doplot) then
             select case (ccdN)
             case (2)
                write(lunqdp,2120) rate_DP(i), urate_DP(i), &
                     hard_DP(i),uhard_DP(i), d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2
2120            format(1x,2f10.6,2f6.2,34f4.0)
             case (3)
                write(lunqdp,2130) rate_DP(i), urate_DP(i), &
                     d1, d2, d1, d2, d1, d2,  &
                     hard_DP(i), uhard_DP(i), d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2
2130            format(1x,2f10.6,6f4.0,2f6.2,28f4.0)
             case (4)
                write(lunqdp,2140) rate_DP(i), urate_DP(i), &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2,  &
                     hard_DP(i), uhard_DP(i), d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2
2140            format(1x,2f10.6,12f4.0,2f6.2,22f4.0)
             case (5)
                write(lunqdp,2150) rate_DP(i), urate_DP(i), &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, hard_DP(i), uhard_DP(i), d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2
2150            format(1x,2f10.6,18f4.0,2f6.2,16f4.0)
             case (6)
                write(lunqdp,2160) rate_DP(i), urate_DP(i), &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     hard_DP(i), uhard_DP(i), d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2
2160            format(1x,2f10.6,24f4.0,2f6.2,10f4.0)
             case (7)
                write(lunqdp,2170) rate_DP(i), urate_DP(i), &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, &
                     hard_DP(i), uhard_DP(i), d1, d2, d1, d2
2170            format(1x,2f10.6,30f4.0,2f6.2,4f4.0)
             case default ! BHP: should NEVER happen, but it did, once. 
                call error ("noCCDN",&
                     "aug_get_exp: ccd is none of: [2-7]")
             end select
          endif
          !
          ! SS: Check to see whether the observation CCD is in an odd state
          !     and select whether the data will be kept or not
          !
          ikeep = 1
          if(mosnum .eq. 1) then
             if (ccdN .eq. 4) then
                tr = 100.0 * ratein ! c_val(1)
                if(tr .lt. 5.2) then
                   temphard = 0.96
                elseif(tr .lt. 5.2) then
                   temphard = 0.423*tr - 1.24
                else
                   temphard = 0.423*tr - 1.24
                endif
                if(iodd .eq. 1) then
                   if(hard_DP(i) .ge. temphard) ikeep = 0
                else
                   if(hard_DP(i) .lt. temphard) ikeep = 0
                endif
             elseif (ccdN .eq. 5) then
                tr = 100.0 * ratein ! OLDNEW: c_val(1)/ratein
                if(tr .lt. 7.0) then
                   temphard = 1.54
                elseif (tr .lt. 9.0) then
                   temphard = 0.53*tr - 2.17
                else
                   temphard = 2.60
                endif
                if(iodd .eq. 1) then
                   if(hard_DP(i) .gt. temphard) ikeep = 0
                else
                   if(hard_DP(i) .lt. temphard) ikeep = 0
                endif
             else
                if(hard_DP(i) .lt. 1.5) then
                   ikeep = 0
                endif
             endif
          endif
          if(mosnum .eq. 2) then
             if(ccdN .eq. 2) then
                if(iodd .eq. 1) then
                   if(hard_DP(i) .ge. 1.1) ikeep = 0
                else
                   if(hard_DP(i) .lt. 1.1) ikeep = 0
                endif
             elseif (ccdN .eq. 5) then
                tr = 100.0 * ratein
                if(tr .lt. 10.0) then
                   temphard = 0.15*tr
                elseif (tr .lt. 16.67) then
                   temphard = 0.15*tr
                else
                   temphard = 2.50
                endif
                if(iodd .eq. 1) then
                   if(hard_DP(i)  .ge. temphard) ikeep = 0
                else
                   if(hard_DP(i)  .lt. temphard) ikeep = 0
                endif
             else
                if (iodd .eq. 1) then
                   if(hard_DP(i)  .ge. 1.5) ikeep = 0
                else
                   if(hard_DP(i)  .lt. 1.5) ikeep = 0
                endif
             endif
          endif
          !
          ! BHP: For all where ikeep=1 transfer the information
          !      The line below is (probably) unnecessary since
          !      it's caught at the start of the do loop. Meh.
          if(index(filt(i+1),'CalClo') .ne. 0) ikeep = 0 
          !
          ! SS: Keep the data set so transfer the information
          !
          if(ikeep .eq. 1) then
             revl_JP(nobs)  = revl_JP(i)
             expo_DP(nobs)  = expo_DP(i)
             rate_DP(nobs)  = rate_DP(i)
             urate_DP(nobs) = urate_DP(i)
             hard_DP(nobs)  = hard_DP(i)
             uhard_DP(nobs) = uhard_DP(i)
             do ii=0,chans-1
                spec_JP(ii,nobs) = spec_JP(ii,i)
             enddo
          endif
          exx = exx + expo_DP(i)
          nobs = nobs + 1
       endif
    end do
    write(errstr,*)' Total exposure for selected obsids= ',exx,' nobs = ',nobs
    call message(AppMsg,Sparse)
    !
    ! SS: Set the energies of the bins
    !
    allocate(energ(chans))
    do i=1,chans ! 2400
       energ(i) = 0.005*(i - 1)
    enddo
    !
    ! SS: Get some temporary values
    !
    murate = 0.0
    muhard = 0.0
    do i=1,nobs
       murate = murate + urate_DP(i)
       muhard = muhard + uhard_DP(i)
    enddo
    if (nobs .ne. 0) then 
       murate = murate/nobs
       muhard = muhard/nobs
    endif
    !
    write(cccd,'(i1)')ccdN
    write(errstr,'(A,f12.5,A,f12.5)')'Rate: ',ratein,' Rate error: ',murate
    call message(AppMsg,Sparse)
    write(errstr,'(A,f12.5,A,f12.5)')'Hard: ',hardin,' Hard error: ',muhard
    call message(AppMsg,Sparse)
    !
    ! SS: Find the closest c_tim secs
    !
    allocate(dist(qpbRows),ddist(qpbRows))
    allocate(hdist(qpbRows),rdist(qpbRows))
    allocate(tdist(qpbRows),ord(qpbRows))
    !
    if(c_tim .eq. 0.0) then
       call error ("NoExpoLen","No exposure length provided")
       !       return ***return gracefully after that???!!!***
    else
       if((c_dst(1) .eq. 0.0) .and. (c_dst(2) .eq. 0.0) .and. &
            (c_dst(3) .eq. 0.0)) then
          call error("NoDistScal","No distance scaling provided")
       else
          do i=1,nobs
             if(c_dst(1) .gt. 0.0) then
                rdist(i) = (rate_DP(i)-ratein)/c_dst(1)
             else
                rdist(i) = rate_DP(i)*0.0
             endif
             if(c_dst(2) .gt. 0.0) then
                hdist(i) = (hard_DP(i)-hardin)/c_dst(2)
             else
                hdist(i) = hard_DP(i)*0.0
             endif
             if(c_dst(3) .gt. 0.0) then
                ddist(i) = (revl_JP(i)-revlin)/c_dst(3)
             else
                ddist(i) = revl_JP(i)*0.0
             endif
             dist(i) = sqrt(rdist(i)**2 + hdist(i)**2 + ddist(i)**2)
          enddo
       endif
       do i=1,nobs
          ord(i) = i
          tdist(i) = dist(i)
       enddo
       do ii=nobs-1,1,-1
          do i=1,ii
             if(tdist(i) .gt. tdist(i+1)) then
                td = tdist(i)
                tdist(i) = tdist(i+1)
                tdist(i+1) = td
                itd = ord(i)
                ord(i) = ord(i+1)
                ord(i+1) = itd
             endif
          enddo
       enddo
       cum_exp = 0.0
       n = 0
       ave_rate = 0.0
       modd = 0 ! fix this ???!!!
       do while ((cum_exp .lt. c_tim) .and. (n .lt. nobs))
          n = n + 1
          cum_exp = cum_exp + expo_DP(ord(n))
          ave_rate = ave_rate + rate_DP(ord(n))*expo_DP(ord(n))
          if(doplot) then
             !             modd = mod(n,4) ! print every 4th line 'too many lines' QDP error???!!!
             write(cccd,'(I1)')ccdN
             write(errstr,*)"Writing CCD"//cccd//" to QDP file"
             call message(AppMsg,Noisy)
             !             modd = mod(n,4)
             if (modd .eq. 0) then
                select case (ccdN)
                case (2)
                   write(lunqdp,2125) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, hard_DP(ord(n)), uhard_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2
2125               format(1x,2f10.6,2f4.0,2f6.2,32f4.0)
                case (3)
                   write(lunqdp,2135) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2,  &
                        hard_DP(ord(n)), uhard_DP(ord(n)), d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2,  &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2
2135               format(1x,2f10.6,8f4.0,2f6.2,26f4.0)
                case (4)
                   write(lunqdp,2145) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, &
                        hard_DP(ord(n)), uhard_DP(ord(n)),  &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2
2145               format(1x,2f10.6,14f4.0,2f6.2,20f4.0)
                case (5)
                   write(lunqdp,2155) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, &
                        hard_DP(ord(n)), uhard_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, & 
                        d1, d2, d1, d2, d1, d2
2155               format(1x,2f10.6,20f4.0,2f6.2,14f4.0)
                case (6)
                   write(lunqdp,2165) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, hard_DP(ord(n)), uhard_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2
2165               format(1x,2f10.6,26f4.0,2f6.2,8f4.0)
                case (7)
                   write(lunqdp,2175) rate_DP(ord(n)), urate_DP(ord(n)), &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2, &
                        d1, d2, d1, d2, d1, d2, d1, d2,  &
                        d1, d2, d1, d2, d1, d2, d1, d2,  &
                        hard_DP(ord(n)), uhard_DP(ord(n)), d1, d2
2175               format(1x,2f10.6,32f4.0,2f6.2,2f4.0)
                case default ! should NEVER happen! 
                   call error ("noCCDN",&
                        "aug_get_exp: ccd is none of: [2-7]")
                end select
             end if
          endif
       enddo
       if (cum_exp .ne. 0) then 
          ave_rate = ave_rate/cum_exp
       else
          call error("expLEzero","Augmented spectrum cumulative exposure LE 0.")
       end if
       write(errstr,4010) 'Exposure: ', cum_exp
       call message(AppMsg,Always)
4010   format(A,f10.2)
       write(errstr,*) 'Exposure weighted count rate: ', ave_rate
       call message(AppMsg,Always)
       write(errstr,4020)  'Observation count rate: ', ratein, &
            ' and ratio: ', ave_rate/ratein
       call message(AppMsg,Always)
4020   format(A,f12.4,A,f10.6)
       ! BHP: p_flg hardcoded to 1 so if statement removed
       if (doplot) then
          if (modd .eq. 0) then 
             if(ccdN .eq. 2) then
                write(lunqdp,2127) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, hardin, 0.2*hardin, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2
2127            format(1x,2f10.6,4f4.0,2f6.2,30f4.0)
             elseif(ccdN .eq. 3) then
                write(lunqdp,2137) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, d1, d2, d1, d2,  &
                     d1, d2, hardin, 0.2*hardin, &
                     d1, d2, d1, d2, d1, d2, d1, d2,  &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2
2137            format(1x,2f10.6,10f4.0,2f6.2,24f4.0)
             elseif(ccdN .eq. 4) then
                write(lunqdp,2147) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     hardin, 0.2*hardin,  &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2
2147            format(1x,2f10.6,16f4.0,2f6.2,18f4.0)
             elseif(ccdN .eq. 5) then
                write(lunqdp,2157) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2,  &
                     hardin, 0.2*hardin, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2
2157            format(1x,2f10.6,22f4.0,2f6.2,12f4.0)
             elseif(ccdN .eq. 6) then
                write(lunqdp,2167) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, hardin, 0.2*hardin, &
                     d1, d2, d1, d2, d1, d2
2167            format(1x,2f10.6,28f4.0,2f6.2,6f4.0)
             elseif(ccdN .eq. 7) then
                write(lunqdp,2177) ratein, 0.2*ratein, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2, &
                     d1, d2, d1, d2, d1, d2, d1, d2,  &
                     d1, d2, d1, d2, d1, d2, d1, d2,  &
                     d1, d2, hardin, 0.2*hardin
2177            format(1x,2f10.6,34f4.0,2f6.2)
             endif
          end if
       endif
    end if
    !
    ! SS: Actually put together the spectrum
    !
    aug_spec(ii) = 0.0 ! 2400 channel array
    aug_expo = 0.
    !
    do i=1,n
       do ii=1,chans
          aug_spec(ii) = aug_spec(ii) + real(spec_JP(ii-1,ord(i)))
       enddo
       aug_expo = aug_expo + expo_DP(ord(i))
    enddo
    aug_expo = aug_expo*ave_rate/ratein
    write(errstr,4020)  'Augmented exposure (augexp*averate/ratein): ',aug_expo
    call message(AppMsg,Always)
    call release (QPB_T)

    return

  end subroutine augment_mos
  !
  ! was aug_get_pnexp in ESAS
  !
  subroutine augment_pn (QPB_S,ratein,hardin,revlin,ootsca,c_dst,c_tim,&
       aug_spec,aug_expo,quadN,chans,obsid,expid,smodein,plc_lo,plc_hi,&
       lunaug,lunspec,lunrr,lunrh,firsttime,doplot,pattern)
    ! OLD: subroutine aug_get_pnexp (db_file,c_val,c_dst,c_tim,ichip,&
    !           obsid,prefix,modeint,plc_lo,plc_hi)
    !
    ! called by ESAS routine(s):
    !  pnback (OLD: pn_back)
    !
    ! Inputs: fix this 
    !   qdp_F: QDP plot file name (pnS001-aug.qdp)
    !   QPB_S: quiescent particle background CCF set already opened
    !   ratein: R*4 rate from pnback was c_val(1)
    !   hardin: R*4 hard from pnback was c_val(2)
    !   revlin: I*4 revl from pnback was c_val(3)
    !   ootsca: R*4 OOT scalefactor, derived in pnback
    !   c_dst(3) (/0.01,1.,0.1/) hardcoded in pnback R*4 aug_scal
    !   c_tim    1E06 hardcoded in pnback            R*4 aug_tim
    !   aug_spec(chans) OUT augmented spectra R*4
    !   aug_expo        OUT augmented exposure R*4
    !   quadN           2-7    I*4
    !   chans           4096 for pn, variable I*4
    !   obsid           1234567890 C*10
    !   expid           S002      C*5
    !   smodein         PN SUBMODE
    !   plc_lo(2)       Low channel energy range from pnback R*4
    !   plc_hi(2)       Hi channel energy range from pnback R*4
    !   lunqug          Logical unit number for output aug spec QDP file I*4
    !   lunspec         Logical unit number for output spec QDP file I*4
    !   lunrr           Logical unit number for output rev rate QDP file I*4
    !   lunrh           Logical unit number for output hard rev rate QDP file I*4
    !   firsttime       first time thru here? L*1 for opening QDP file
    !   doplot   usually 'withqdp' logical  L*1
    !   luns go in and out again for each quad selected
    type(DataSetT), intent(inout) :: QPB_S
    type(TableT)   :: QPB_T
    type(ColumnT)  :: obsid_C, expid_C, date_C, revl_C, filter_C, expo_C, &
         submode_C, rate_C, urate_C, hard_C, uhard_C, rate1_C, urate1_C, &
         hard1_C, uhard1_C, rate2_C, urate2_C, hard2_C, uhard2_C, &
         rate3_C, urate3_C, hard3_C, uhard3_C, rate4_C, urate4_C, &
         ttime_C, spec_C, specoot_C, back_C, quad_C
    !
    character(len=*), intent(in) :: expid, obsid, smodein
    character(len=256) :: qdp_F, QPB_F, qdpaug_F, qdpspec_F, qdprr_F, qdprh_F, &
         expression
    character(len=25), dimension(:), allocatable :: filt, submode
    character(len=24) :: title
    character(len=20) :: extname
    character(len=12) :: cexpmax, cexpmin, cratemin, cratemax, chardmin, chardmax
    character(len=1) :: cquad
    integer, intent(in) :: pattern, chans, quadN, revlin, plc_lo(2), plc_hi(2)
    integer, intent(inout) :: lunaug, lunspec, lunrr, lunrh
    integer(kind=int16), allocatable, dimension(:) :: quad
    integer(kind=int16), dimension(:), pointer :: quad_IP
    integer(kind=int32), dimension(:,:), pointer :: spec_JP, specoot_JP
    integer(kind=int32), dimension(:), pointer :: revl_JP, back_JP
    integer :: nRows1, isWrong, nobs, modd, datatype
    integer :: i, ii, n, nn, ikeep, iodd, itd, nchan, goodRows, qpbRows
    integer :: nb1, nb2, nb3, ng, ntot, status, totRows, nChans = 4096
    integer, allocatable, dimension(:) :: revl, ord
    logical, intent(inout) :: doplot, firsttime
    logical :: warndie
    real(kind=single), intent(in) :: c_tim, c_dst(3), ratein, hardin, ootsca
    real(kind=single), intent(out) :: aug_expo
    real(kind=single) :: amscal = 1440000.0
    real(kind=single) :: ave_rate, cum_exp, d1, d2, hardmax, hardmin, &
         muhard, murate, ratemax, ratemin, td, temphard, temp, tr, tootsca, &
         te1, te2, teo1, teo2, explo, exphi
    real(kind=single) :: aug_spec(chans) ! 4096
    real(kind=single), allocatable, dimension(:,:) :: specBig, specBigOOT ! (4096,15000)
    real(kind=single), allocatable, dimension(:) :: back, ddist, dist, &
         energ, expo, hard, hard2, hdist, rate, rdist, tdist, time, allootsca, &
         uhard, urate, uhard2, ttime
    real(kind=single), dimension(:), pointer   :: expo_RP
    real(kind=single), dimension(:,:), pointer :: spec_RP, specoot_RP
    real(kind=double), dimension(:), pointer   :: expo_DP, date_DP, augspec_RP
    real(kind=double), dimension(:), pointer   :: rate_DP, urate_DP, &
         hard_DP, uhard_DP, rate1_DP, urate1_DP, hard1_DP, uhard1_DP, &
         rate2_DP, urate2_DP, hard2_DP, uhard2_DP, rate3_DP, urate3_DP, &
         hard3_DP, uhard3_DP, rate4_DP, urate4_DP, ttime_DP
    !
    status = 0
    write(errstr,4001) ratein,hardin,revlin ! OLD: c_val(1:3)
    call message(AppMsg,Always)
4001 format('Count Rate, Hardness, and Revolution: ',f8.6,f8.3,i5)
    write(errstr,4002) c_dst ! OLD: c_dst
    call message(AppMsg,Always)
4002 format('Scaling Factors for Obs. Selection  : ',f8.6,f8.3,f8.0)
    write(errstr,4003)ootsca    
    call message(AppMsg,Always)
4003 format('Observation OOT Scale Factor        : ',f7.4)
    !
    ! qdpfile was augfile
    ! create title for qdp plot
    !
    write(cquad,'(I1)') quadN
    title = '"' // adjustl(trim(obsid)) // ' -- PN ' // adjustl(trim(expid)) // '"'
    !
    ! If selected, make a plot, first set up the plot
    ! One of four: qdpaug_F, qdpspec_F, qdprr_F, qdprh_F
    !
    if(doplot) then
       if(firsttime) then 
          write(errstr,*)"Starting four QDP plots: ratehard,augspec,rrate,rhard"
          call message(AppMsg,Always)
          !
          ! QDP file 1 augmented spectrum QDP plot
          ! rate v hardness 
          ! OLDNEW: pnS001-aug.qdp/pnS001-ratehard.qdp
          !
          qdpaug_F = "pn" // adjustl(trim(expid)) // "-ratehard.qdp" 
          !
          ! already check_clobber in pnback, just need returnable lun
          call open_ascii_file(qdpaug_F,lunaug) ! OLD: augfile, lunp
          write(lunaug,*) ' viewport 0.2 0.2 0.9 0.8'
          write(lunaug,*) ' la t ',adjustl(trim(title)) ! created at top
          write(lunaug,*) ' read serr 1'
          write(lunaug,*) ' read serr 2'
          write(lunaug,*) ' read serr 3'
          write(lunaug,*) ' read serr 4'
          write(lunaug,*) ' read serr 5'
          write(lunaug,*) ' read serr 6'
          write(lunaug,*) ' read serr 7'
          write(lunaug,*) ' read serr 8'
          write(lunaug,*) ' read serr 9'
          write(lunaug,*) ' font roman'
          write(lunaug,*) ' cs 1.5'
          write(lunaug,*) ' la f'
          write(lunaug,*) ' col 2 on 2'
          write(lunaug,*) ' col 1 on 3'
          write(lunaug,*) ' col 2 on 4'
          write(lunaug,*) ' col 1 on 5'
          write(lunaug,*) ' col 2 on 6'
          write(lunaug,*) ' col 1 on 7'
          write(lunaug,*) ' col 2 on 8'
          write(lunaug,*) ' col 1 on 9'
          write(lunaug,*) ' mark 1 on 2'
          write(lunaug,*) ' mark 1 on 3'
          write(lunaug,*) ' mark 1 on 4'
          write(lunaug,*) ' mark 1 on 5'
          write(lunaug,*) ' mark 1 on 6'
          write(lunaug,*) ' mark 1 on 7'
          write(lunaug,*) ' mark 1 on 8'
          write(lunaug,*) ' mark 1 on 9'
          write(lunaug,*) ' line off 2'
          write(lunaug,*) ' line off 3'
          write(lunaug,*) ' line off 4'
          write(lunaug,*) ' line off 5'
          write(lunaug,*) ' line off 6'
          write(lunaug,*) ' line off 7'
          write(lunaug,*) ' line off 8'
          write(lunaug,*) ' line off 9'
          write(lunaug,*) ' win 1'
          write(lunaug,*) ' yplot 1 2 3'
          write(lunaug,*) ' loc 0.05 0.45 0.57 1.0'
          write(lunaug,*) ' la 11 p 0.00065 2.6 just left "Quad #1"'
          write(lunaug,*) ' r x1 0.0005 0.004'
          write(lunaug,*) ' r y1 2.0 8.0'
          write(lunaug,*) ' la y Hardness'
          write(lunaug,*) ' win 2'
          write(lunaug,*) ' yplot 1 4 5'
          write(lunaug,*) ' loc 0.47 0.45 0.99 1.0'
          write(lunaug,*) ' la 12 p 0.00065 2.6 just left "Quad #2"'
          write(lunaug,*) ' r x2 0.0005 0.004'
          write(lunaug,*) ' r y2 2.0 8.0'
          write(lunaug,*) ' win 3'
          write(lunaug,*) ' yplot 1 6 7'
          write(lunaug,*) ' loc 0.05 0.05 0.57 0.60'
          write(lunaug,*) ' la 13 p 0.00065 2.6 just left "Quad #3"'
          write(lunaug,*) ' r x3 0.0005 0.004'
          write(lunaug,*) ' r y3 2.0 8.0'
          write(lunaug,*) ' la y Hardness'
          write(lunaug,*) ' la x Rate'
          write(lunaug,*) ' win 4'
          write(lunaug,*) ' yplot 1 8 9'
          write(lunaug,*) ' loc 0.47 0.05 0.99 0.60'
          write(lunaug,*) ' la 14 p 0.00065 2.6 just left "Quad #4"'
          write(lunaug,*) ' r x4 0.0005 0.004'
          write(lunaug,*) ' r y4 2.0 8.0'
          write(lunaug,*) ' col off 1'
          write(lunaug,*) ' la x Rate'
          !
          ! QDP file 2 augmented spectrum QDP plot
          ! SS: Get another unused lun for the augmented spectra
          ! OLDNEW: pnS001-aug-spec.qdp/pnS001-augspec.qdp
          qdpspec_F = "pn" // adjustl(trim(expid)) // "-augspec.qdp"
          ! check_clobber done in pnback, only need lun
          ! create title for qdp plot
          call open_ascii_file(qdpspec_F,lunspec) ! OLD: augfile, lunpp
          write(lunspec,*) ' viewport 0.2 0.2 0.9 0.8'
          write(lunspec,*) ' la t ', adjustl(trim(title))
          write(lunspec,*) ' font roman'
          write(lunspec,*) ' cs 1.5'
          write(lunspec,*) ' la f'
          write(lunspec,*) ' col 1 on 2'
          write(lunspec,*) ' col 1 on 3'
          write(lunspec,*) ' col 1 on 4'
          write(lunspec,*) ' col 1 on 5'
          write(lunspec,*) ' mark 1 on 2'
          write(lunspec,*) ' mark 1 on 3'
          write(lunspec,*) ' mark 1 on 4'
          write(lunspec,*) ' mark 1 on 5'
          write(lunspec,*) ' line off 2'
          write(lunspec,*) ' line off 3'
          write(lunspec,*) ' line off 4'
          write(lunspec,*) ' line off 5'
          write(lunspec,*) ' win 1'
          write(lunspec,*) ' yplot 1 2'
          write(lunspec,*) ' loc 0.05 0.45 0.57 1.0'
          write(lunspec,*) ' la 11 p 4.0 200.0 just left "Quad #1"'
          write(lunspec,*) ' r x1 0.15 15.0'
          write(lunspec,*) ' r y1 0.1 500.0'
          write(lunspec,*) ' la y Counts'
          write(lunspec,*) ' log x on'
          write(lunspec,*) ' log y on'
          write(lunspec,*) ' win 2'
          write(lunspec,*) ' yplot 1 3'
          write(lunspec,*) ' loc 0.47 0.45 0.99 1.0'
          write(lunspec,*) ' la 12 p 4.0 200.0 just left "Quad #2"'
          write(lunspec,*) ' r x2 0.15 15.0'
          write(lunspec,*) ' r y2 0.1 500.0'
          write(lunspec,*) ' log x on'
          write(lunspec,*) ' log y on'
          write(lunspec,*) ' win 3'
          write(lunspec,*) ' yplot 1 4'
          write(lunspec,*) ' loc 0.05 0.05 0.57 0.60'
          write(lunspec,*) ' la 13 p 4.0 200.0 just left "Quad #3"'
          write(lunspec,*) ' r x3 0.15 15.0'
          write(lunspec,*) ' r y3 0.1 500.0'
          write(lunspec,*) ' la y Counts'
          write(lunspec,*) ' la x Energy (keV)'
          write(lunspec,*) ' log x on'
          write(lunspec,*) ' log y on'
          write(lunspec,*) ' win 4'
          write(lunspec,*) ' yplot 1 5'
          write(lunspec,*) ' loc 0.47 0.05 0.99 0.60'
          write(lunspec,*) ' la 14 p 4.0 200.0 just left "Quad #4"'
          write(lunspec,*) ' r x4 0.15 15.0'
          write(lunspec,*) ' r y4 0.1 500.0'
          write(lunspec,*) ' col off 1'
          write(lunspec,*) ' la x Energy (keV)'
          write(lunspec,*) ' log x on'
          write(lunspec,*) ' log y on'
          !
          ! QDP file 3 augmented rev rate QDP plot
          ! filename: pnS001-augrrate.qdp
          !
          ! SS: Get another unused lun for the selected rate 
          !  plotted with revolution
          ! OLDNEW: pnS001-aug-rev-rate.qdp/pnS001-augraterev.qdp
          qdprr_F = "pn" // adjustl(trim(expid)) // "-augraterev.qdp"
          ! check_clobber done in pnback
          call open_ascii_file(qdprr_F,lunrr) ! OLD: lunppp
          write(lunrr,*) ' viewport 0.2 0.2 0.9 0.8'
          write(lunrr,*) ' la t ', adjustl(trim(title))
          write(lunrr,*) ' read serr 2'
          write(lunrr,*) ' read serr 3'
          write(lunrr,*) ' read serr 4'
          write(lunrr,*) ' read serr 5'
          write(lunrr,*) ' read serr 6'
          write(lunrr,*) ' read serr 7'
          write(lunrr,*) ' read serr 8'
          write(lunrr,*) ' read serr 9'
          write(lunrr,*) ' font roman'
          write(lunrr,*) ' cs 1.5'
          write(lunrr,*) ' la f'
          write(lunrr,*) ' col 1 on 2'
          write(lunrr,*) ' col 2 on 3'
          write(lunrr,*) ' col 1 on 4'
          write(lunrr,*) ' col 3 on 5'
          write(lunrr,*) ' col 1 on 6'
          write(lunrr,*) ' col 4 on 7'
          write(lunrr,*) ' col 1 on 8'
          write(lunrr,*) ' col 5 on 9'
          write(lunrr,*) ' mark 1 on 2'
          write(lunrr,*) ' mark 1 on 3'
          write(lunrr,*) ' mark 1 on 4'
          write(lunrr,*) ' mark 1 on 5'
          write(lunrr,*) ' mark 1 on 6'
          write(lunrr,*) ' mark 1 on 7'
          write(lunrr,*) ' mark 1 on 8'
          write(lunrr,*) ' mark 1 on 9'
          write(lunrr,*) ' line off 2'
          write(lunrr,*) ' line off 3'
          write(lunrr,*) ' line off 4'
          write(lunrr,*) ' line off 5'
          write(lunrr,*) ' line off 6'
          write(lunrr,*) ' line off 7'
          write(lunrr,*) ' line off 8'
          write(lunrr,*) ' line off 9'
          write(lunrr,*) ' win 1'
          write(lunrr,*) ' yplot 1 2 3'
          write(lunrr,*) ' loc 0.05 0.45 0.57 1.0'
          write(lunrr,*) ' la 11 p 200.0 0.0036 just left "Quad #1"'
          write(lunrr,*) ' r x1 0.0 3500.0'
          write(lunrr,*) ' r y1 0.001 0.004'
          write(lunrr,*) ' la y Count Rate'
          write(lunrr,*) ' win 2'
          write(lunrr,*) ' yplot 1 4 5'
          write(lunrr,*) ' loc 0.47 0.45 0.99 1.0'
          write(lunrr,*) ' la 12 p 200.0 0.0036 just left "Quad #2"'
          write(lunrr,*) ' r x2 0.0 3500.0'
          write(lunrr,*) ' r y2 0.001 0.004'
          write(lunrr,*) ' win 3'
          write(lunrr,*) ' yplot 1 6 7'
          write(lunrr,*) ' loc 0.05 0.05 0.57 0.60'
          write(lunrr,*) ' la 13 p 200.0 0.0036 just left "Quad #3"'
          write(lunrr,*) ' r x3 0.0 3500.0'
          write(lunrr,*) ' r y3 0.001 0.004'
          write(lunrr,*) ' la y Count Rate'
          write(lunrr,*) ' la x Revolution'
          write(lunrr,*) ' win 4'
          write(lunrr,*) ' yplot 1 8 9'
          write(lunrr,*) ' loc 0.47 0.05 0.99 0.60'
          write(lunrr,*) ' la 14 p 200.0 0.0036 just left "Quad #4"'
          write(lunrr,*) ' r x4 0.0 3500.0'
          write(lunrr,*) ' r y4 0.001 0.004'
          write(lunrr,*) ' col off 1'
          write(lunrr,*) ' la x Revolution'
          !
          ! QDP file 4 augmented rev hard QDP plot
          !
          ! SS: Get another unused lun for the selected rate 
          !  plotted with revolution
          ! OLDNEW: pnS001-aug-rev-hard.qdp/pnS001-augratehrev.qdp
          qdprh_F = "pn" // adjustl(trim(expid)) // "-augratehrev.qdp"
          ! check_clobber done in pnback
          call open_ascii_file(qdprh_F,lunrh) ! OLD: lunpppp
          write(lunrh,*) ' viewport 0.2 0.2 0.9 0.8'
          write(lunrh,*) ' la t ', adjustl(trim(title))
          write(lunrh,*) ' read serr 2'
          write(lunrh,*) ' read serr 3'
          write(lunrh,*) ' read serr 4'
          write(lunrh,*) ' read serr 5'
          write(lunrh,*) ' read serr 6'
          write(lunrh,*) ' read serr 7'
          write(lunrh,*) ' read serr 8'
          write(lunrh,*) ' read serr 9'
          write(lunrh,*) ' font roman'
          write(lunrh,*) ' cs 1.5'
          write(lunrh,*) ' la f'
          write(lunrh,*) ' col 1 on 2'
          write(lunrh,*) ' col 2 on 3'
          write(lunrh,*) ' col 1 on 4'
          write(lunrh,*) ' col 3 on 5'
          write(lunrh,*) ' col 1 on 6'
          write(lunrh,*) ' col 4 on 7'
          write(lunrh,*) ' col 1 on 8'
          write(lunrh,*) ' col 5 on 9'
          write(lunrh,*) ' mark 1 on 2'
          write(lunrh,*) ' mark 1 on 3'
          write(lunrh,*) ' mark 1 on 4'
          write(lunrh,*) ' mark 1 on 5'
          write(lunrh,*) ' mark 1 on 6'
          write(lunrh,*) ' mark 1 on 7'
          write(lunrh,*) ' mark 1 on 8'
          write(lunrh,*) ' mark 1 on 9'
          write(lunrh,*) ' line off 2'
          write(lunrh,*) ' line off 3'
          write(lunrh,*) ' line off 4'
          write(lunrh,*) ' line off 5'
          write(lunrh,*) ' line off 6'
          write(lunrh,*) ' line off 7'
          write(lunrh,*) ' line off 8'
          write(lunrh,*) ' line off 9'
          write(lunrh,*) ' win 1'
          write(lunrh,*) ' yplot 1 2 3'
          write(lunrh,*) ' loc 0.05 0.45 0.57 1.0'
          write(lunrh,*) ' la 11 p 200.0 9.0 just left "Quad #1"'
          write(lunrh,*) ' r x1 0.0 3500.0'
          write(lunrh,*) ' r y1 0.0 10.0'
          write(lunrh,*) ' la y Hardness'
          write(lunrh,*) ' win 2'
          write(lunrh,*) ' yplot 1 4 5'
          write(lunrh,*) ' loc 0.47 0.45 0.99 1.0'
          write(lunrh,*) ' la 12 p 200.0 9.0 just left "Quad #2"'
          write(lunrh,*) ' r x2 0.0 3500.0'
          write(lunrh,*) ' r y2 0.0 10.0'
          write(lunrh,*) ' win 3'
          write(lunrh,*) ' yplot 1 6 7'
          write(lunrh,*) ' loc 0.05 0.05 0.57 0.60'
          write(lunrh,*) ' la 13 p 200.0 9.0 just left "Quad #3"'
          write(lunrh,*) ' r x3 0.0 3500.0'
          write(lunrh,*) ' r y3 0.0 10.0'
          write(lunrh,*) ' la y Hardness'
          write(lunrh,*) ' la x Revolution'
          write(lunrh,*) ' win 4'
          write(lunrh,*) ' yplot 1 8 9'
          write(lunrh,*) ' loc 0.47 0.05 0.99 0.60'
          write(lunrh,*) ' la 14 p 200.0 9.0 just left "Quad #4"'
          write(lunrh,*) ' r x4 0.0 3500.0'
          write(lunrh,*) ' r y4 0.0 10.0'
          write(lunrh,*) ' col off 1'
          write(lunrh,*) ' la x Revolution'
       endif
    endif
    !  SS: Get an unused lun for the CalDB file
    ! BHP: Extract all the relevant columns from the QPB CCF
    ! 
    extname = "QPB" ! or...???!!!
    if (.not.hasBlock(QPB_S,trim(adjustl(extname)))) then ! should never happen!!!
       call error("noQPB", "CAL QPB file does not contain QPB extension")
    else
       QPB_T = table(QPB_S,trim(adjustl(extname))) ! read the whole extension
       qpbRows = numberOfRows(QPB_T)               ! get number of rows
       write(errstr,*)"CAL QPB file has ",qpbRows, " rows"
       call message(AppMsg,Always) ! change this ???!!!
    endif
    !
    !*** BHP: now read the columns in testprods/esas_data/epn_qpb.fits.gz ***
    !*** Columns = obsid, name, object, date, revolut, filter, ontime,    ***
    !*** expo, rate, rate_err, hard, hard_err, etc, spectrum, spectrumOOT ***
    ! The old aug_get_pnexp looped through the data row by row. The DAL
    ! puts every column into a data pointer that can be accessed as an array
    ! what a about KEY column for SAS-21 filter on KEY???!!!
    ! Or we can prefilter with makeqpb???!!!
    !
    write(errstr,*)"Reading relevant columns from QPB SPECTRUM extension"
    ! Commented out columns were never actually USED in the old ESAS
    ! But they remain for any possible future usage (just uncomment them!)
    !    if (.not.hasColumn(QPB_T,"DATE")) then
    !       call error("badColumn","Input file incorrect - no DATE column")
    !    else
    !       date_C = column(QPB_T,"DATE",READ)
    !       date_DP => real64Data(date_C)
    !    endif
    if (.not.hasColumn(QPB_T,"QUAD")) then
       call error("badColumn","Input file incorrect - no QUAD column")
    else
       quad_C = column(QPB_T,"QUAD",READ)
       quad_IP => int16Data(quad_C)
    endif
    if (.not.hasColumn(QPB_T,"REVOLUT")) then
       call error("badColumn","Input file incorrect - no REVOLUT column")
    else
       revl_C = column(QPB_T,"REVOLUT",READ)
       revl_JP => int32Data(revl_C)
    endif
    if (.not.hasColumn(QPB_T,"BACKSCAL")) then
       call error("badColumn","Input file incorrect - no BACKSCAL column")
    else
       back_C = column(QPB_T,"BACKSCAL",READ)
       back_JP => int32Data(back_C)
    endif
    if (.not.hasColumn(QPB_T,"SUBMODE")) then ! filter out more submodes???!!!
       call error("badColumn","Input file incorrect - no SUBMODE column")
    else
       allocate(allootsca(qpbRows))
       allocate(submode(qpbRows))
       submode_C = column(QPB_T,"SUBMODE",READ)
       do i=1,qpbRows
          submode(i) = ""
          submode(i) = stringCell(submode_C,i-1) ! fill chans array with keys
          isWrong = index(submode(i),'Extended') ! fix this???!!!
          if (isWrong .eq. 0) then
             submode(i) = "PrimeFullWindow          "
          else
             submode(i) = "PrimeFullWindowExtended  "
          endif
          !  Set the OOT scale factor for each row
          !
          submode(i) = trim(adjustl(submode(i)))
          call getootsca(submode(i),allootsca(i),.TRUE.) ! in gofutils in goflib
       end do
    endif
    !    if (.not.hasColumn(QPB_T,"FILTER")) then
    !       call error("badColumn","Input file incorrect - no FILTER column")
    !    else
    !       filter_C = column(QPB_T,"FILTER",READ)
    !       allocate(filt(qpbRows))
    !       do i=1,qpbRows
    !          filt(i) = stringCell(filter_C,i-1)  ! fill chans array with keys
    !       end do
    !    endif    
    if (.not.hasColumn(QPB_T,"EXPOSURE")) then
       call error("badColumn","Input file incorrect - no EXPO column")
    else
       expo_C = column(QPB_T,"EXPOSURE",READ)
       datatype = columnDataType( expo_C )
       ! temp fix until EXPOSURE column is corrected in epn_qpb.fits...or just keep???!!!
       if (datatype == REAL32) then
          expo_RP => Real32Data(expo_C)
       else
          expo_DP => Real64Data(expo_C)
       endif
    endif
    !if (.not.hasColumn(QPB_T,"CLEAN_T")) then
    !   call error("badColumn","Input file incorrect - no CLEAN_T column")
    !else
    !   expo_C = column(QPB_T,"CLEAN_T",READ)
    !   expo_DP => Real64Data(expo_C)
    !endif
    !
    if (.not.hasColumn(QPB_T,"RATE")) then
       call error("badColumn","Input file incorrect - no RATE column")
    else
       rate_C = column(QPB_T,"RATE",READ)
       rate_DP => Real64Data(rate_C)
    endif
    if (.not.hasColumn(QPB_T,"RATEERR")) then
       call error("badColumn","Input file incorrect - no RATEERR column")
    else
       urate_C = column(QPB_T,"RATEERR",READ)
       urate_DP => Real64Data(urate_C)
    endif
    !
    !    if (.not.hasColumn(QPB_T,"RATE1")) then
    !       call error("badColumn","Input file incorrect - no RATE1 column")
    !    else
    !       rate1_C = column(QPB_T,"RATE1",READ)
    !       rate1_DP => Real64Data(rate1_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"RATE1ERR")) then
    !       call error("badColumn","Input file incorrect - no RATE1ERR column")
    !    else
    !       urate1_C = column(QPB_T,"RATE1ERR",READ)
    !       urate1_DP => Real64Data(urate1_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"HARD1")) then
    !       call error("badColumn","Input file incorrect - no HARD1 column")
    !    else
    !       hard1_C = column(QPB_T,"HARD1",READ)
    !       hard1_DP => Real64Data(hard1_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"HARD1ERR")) then
    !       call error("badColumn","Input file incorrect - no HARD1ERR column")
    !    else
    !       uhard1_C = column(QPB_T,"HARD1ERR",READ)
    !       uhard1_DP => Real64Data(uhard1_C)
    !    endif
    !
    !    if (.not.hasColumn(QPB_T,"RATE2")) then
    !       call error("badColumn","Input file incorrect - no RATE2 column")
    !    else
    !       rate2_C = column(QPB_T,"RATE2",READ)
    !       rate2_DP => Real64Data(rate2_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"RATE2ERR")) then
    !       call error("badColumn","Input file incorrect - no RATE2ERR column")
    !    else
    !       urate2_C = column(QPB_T,"RATE2ERR",READ)
    !       urate2_DP => Real64Data(urate2_C)
    !    endif
    if (.not.hasColumn(QPB_T,"HARD2")) then
       call error("badColumn","Input file incorrect - no HARD2 column")
    else
       hard2_C = column(QPB_T,"HARD2",READ)
       hard2_DP => Real64Data(hard2_C)
    endif
    if (.not.hasColumn(QPB_T,"HARD2ERR")) then
       call error("badColumn","Input file incorrect - no HARD2ERR column")
    else
       uhard2_C = column(QPB_T,"HARD2ERR",READ)
       uhard2_DP => Real64Data(uhard2_C)
    endif
    !
    !    if (.not.hasColumn(QPB_T,"RATE3")) then
    !       call error("badColumn","Input file incorrect - no RATE3 column")
    !    else
    !       rate3_C = column(QPB_T,"RATE3",READ)
    !       rate3_DP => Real64Data(rate3_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"RATE3ERR")) then
    !       call error("badColumn","Input file incorrect - no RATE3ERR column")
    !    else
    !       urate3_C = column(QPB_T,"RATE3ERR",READ)
    !       urate3_DP => Real64Data(urate3_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"HARD3")) then
    !       call error("badColumn","Input file incorrect - no HARD3 column")
    !    else
    !       hard3_C = column(QPB_T,"HARD3",READ)
    !       hard3_DP => Real64Data(hard3_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"HARD3ERR")) then
    !       call error("badColumn","Input file incorrect - no HARD3ERR column")
    !    else
    !       uhard3_C = column(QPB_T,"HARD3ERR",READ)
    !       uhard3_DP => Real64Data(uhard3_C)
    !    endif
    !
    !    if (.not.hasColumn(QPB_T,"RATE4")) then
    !       call error("badColumn","Input file incorrect - no RATE4 column")
    !    else
    !       rate4_C = column(QPB_T,"RATE4",READ)
    !       rate4_DP => Real64Data(rate4_C)
    !    endif
    !    if (.not.hasColumn(QPB_T,"RATE4ERR")) then
    !       call error("badColumn","Input file incorrect - no RATE4ERR column")
    !    else
    !       urate4_C = column(QPB_T,"RATE4ERR",READ)
    !       urate4_DP => Real64Data(urate4_C)
    !    endif
    if (.not.hasColumn(QPB_T,"TOTAL_T")) then
       call error("badColumn","Input file incorrect - no TOTAL_T column")
    else
       ttime_C = column(QPB_T,"TOTAL_T",READ)
       ttime_DP => Real64Data(ttime_C)
    endif
    allocate(ttime(qpbRows))
    do i=1,qpbRows
       ttime(i) = real(ttime_DP(i-1))
    end do
    nullify(ttime_DP) ! everywhere they're put in used arrays???!!!
    !
    ! spec is (4096,qpbRows) integer 'J' pointer for spec and specoot
    !
    if (pattern .eq. 0) then 
       if (.not.hasColumn(QPB_T,"SPECP0")) then
          call error("badColumn","Input file incorrect - no SPECP0 column")
       else
          spec_C = column(QPB_T,"SPECP0",READ)
          spec_JP => int32VectorData(spec_C)       ! 4096 x qpbRows
       endif
       if (.not.hasColumn(QPB_T,"SPECOOP0")) then
          call error("badColumn","Input file incorrect - no SPECOOP0 column")
       else
          specoot_C = column(QPB_T,"SPECOOP0",READ)
          specoot_JP => int32VectorData(specoot_C) ! 4096 x qpbRows
       endif
    elseif (pattern .eq. 4) then
       if (.not.hasColumn(QPB_T,"SPECP4")) then
          call error("badColumn","Input file incorrect - no SPECP4 column")
       else
          spec_C = column(QPB_T,"SPECP4",READ)
          spec_JP => int32VectorData(spec_C)       ! 4096 x qpbRows
       endif
       if (.not.hasColumn(QPB_T,"SPECOOP4")) then
          call error("badColumn","Input file incorrect - no SPECOOP4 column")
       else
          specoot_C = column(QPB_T,"SPECOOP4",READ)
          specoot_JP => int32VectorData(specoot_C) ! 4096 x qpbRows
       endif
    else
       ! do something else???!!!
    end if
    !
    nn   = 0
    ntot = 0
    ng   = 0
    nb1  = 0
    nb2  = 0
    nb3  = 0
    !
    allocate(specBig(nChans,qpbRows),specBigOOT(nChans,qpbRows))
    allocate(rate(qpbRows),urate(qpbRows),expo(qpbRows),back(qpbRows))
    allocate(hard2(qpbRows),uhard2(qpbRows),revl(qpbRows),quad(qpbRows))
    !
    ! SS: Extract the spectrum and normalize by exposure and area
    ! BHP: check precision???!!!
    nn = 1 ! 'good' row counter
    do i=1,qpbRows ! i='current row' counter
       if (quad_IP(i-1) .eq. quadN) then
          if (datatype == REAL32) then
             expo(nn) = expo_RP(i-1)
          else
             expo(nn) = real(expo_DP(i-1))
          endif
          if (expo(nn) .gt. 100.0) then ! never actually happens!!!
             ! "got a good one..."
             quad(nn)   = quadN ! only the current one, used later
             rate(nn)   = real(rate_DP(i-1))  ! c_tim: 1e+06, passed by pnback,
             urate(nn)  = real(urate_DP(i-1)) ! maybe make param???!!!
             back(nn)   = real(back_JP(i-1))
             revl(nn)   = revl_JP(i-1)
             hard2(nn)  = real(hard2_DP(i-1))
             uhard2(nn) = real(uhard2_DP(i-1))
             te1  = 0.
             do ii=plc_lo(1),plc_lo(2)              ! check indicies???!!!
                te1  = te1  + real(spec_JP(ii-1,i-1)) ! OLD: tspec by-row spec
             enddo
             te2 = 0.
             do ii=plc_hi(1),plc_hi(2) ! check indicies???!!!
                te2  = te2  + real(spec_JP(ii-1,i-1)) ! pointer from DAL
             enddo
             do ii=1,nChans ! 4096 
                specBig(ii,nn) = (amscal * real(spec_JP(ii-1,i-1))) &
                     /(expo(nn) * real(back(nn)))
             end do
             rate(nn) = c_tim * rate(nn) ! c_tim hardcoded to 1E06
             urate(nn) = c_tim * urate(nn)
             ! specBig is good!
             !
             ! SS: Extract the OOT spectrum and normalize by exposure and area
             !
             ! BHP: if(ispecoo .ne. 0) then removed; dies above if column not
             ! in QPB plus, it didn't DO anything if it wasn't there!
             teo1 = 0.0
             do ii=plc_lo(1),plc_lo(2)
                teo1 = teo1 + real(specoot_JP(ii-1,i-1)) ! pointer from DAL
             enddo
             teo2 = 0.0
             do ii=plc_hi(1),plc_hi(2)
                teo2 = teo2 + real(specoot_JP(ii-1,i-1)) ! pointer from DAL
             enddo
             do ii=1,nChans
                if(ii .lt. 600) then
                   temp = 5.0 - 1.3* &
                        (0.005*ii - 0.7)
                elseif(ii .lt. 1400) then
                   temp = 2.0 - 0.125* &
                        (0.005*ii - 3.0)
                else
                   temp = 1.5
                endif
                specBigOOT(ii,nn) = (amscal * real(specoot_JP(ii-1,i-1))*temp) &
                     /(expo(nn) * real(back(nn)))
             enddo
             ! specBig/specBigOOT are good!
             ! SS: Set the OOT scale factor (BHP: already done above)
             ! SS: Full frame extended 5 is well hidden (BHP: taken care of in makeqpb)
             !
             ! SS: Find the normal and OOT counts in the low and high spectral bands
             !
             te1  = 0.
             teo1 = 0.
             do ii=plc_lo(1),plc_lo(2)
                te1  = te1  + specBig(ii,nn)    
                teo1 = teo1 + specBigOOT(ii,nn)
             enddo
             te2  = 0.0
             teo2 = 0.0
             do ii=plc_hi(1),plc_hi(2)
                te2  = te2  + specBig(ii,nn)
                teo2 = teo2 + specBigOOT(ii,nn)
             enddo
             temp = rate(nn) ! rate(nn) from above
             rate(nn) = te1 + te2 - allootsca(i)*(teo1 + teo2) ! changed from above???!!!
             temp = hard2(nn) 
             hard2(nn)  = (te2 - allootsca(i)*teo2)/(te1 - allootsca(i)*teo1)
             nn = nn + 1
          end if
       end if
    end do
    ! Put nullifys here to free memory???!!!
    goodRows = nn - 1 ! why???!!!
    write(errstr,*)"Number of good rows in QPB file = ",goodRows
    call message(AppMsg,Always) ! change this to Noisy???!!!
    !
    ! SS: Set the normal filtering limits
    !
    d1 = -9.
    d2 = 0.
    hardmin = 3.
    hardmax = 7.
    ratemin = 1.0e-3
    ratemax = 3.0e-3
    !
    ! SS: Sort through the archived observation data collecting
    !  the specifics on possible observations that can be used
    !  for the augmentation.
    !
    !  Write out all the data for plotting if selected by the 
    !  rate, exposure, and hardness criteria
    !
    explo = 5000.
    exphi = 40000.
    nobs = 0

    do i=1,goodRows ! subset of qpbRows from above 'nn'
       modd = mod(i,4)
       if (modd .eq. 0) then
          if((expo(i) .gt. 0.0) .and. (ttime(i) .gt. 0.0)) then
             if((expo(i) .ge. explo) .and.  &
                  (expo(i) .lt. exphi) .and. &
                  (rate(i) .gt. ratemin) .and.  &
                  (rate(i) .lt. ratemax) .and.  &
                  (hard2(i) .gt. hardmin) .and. &
                  (hard2(i) .lt. hardmax) .and. &
                  (expo(i) .gt. 0.8*ttime(i))) then
                if(doplot) then
                   if(quadN .eq. 1) then
                      write(lunrh,2820) revl(i), &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2820                  format(1x,i6,2f6.2,14f4.0)
                      write(lunrr,2720) revl(i), &
                           rate(i), urate(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2720                  format(1x,i6,2f10.6,14f4.0)
                      write(lunaug,2120) rate(i), urate(i), &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2120                  format(1x,2f10.6,2f6.2,22f4.0)
                   elseif(quadN .eq. 2) then
                      write(lunrh,2830) revl(i),  &
                           d1, d2, d1, d2,  &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2830                  format(1x,i6,4f4.0,2f6.2,10f4.0)
                      write(lunrr,2730) revl(i),  &
                           d1, d2, d1, d2,  &
                           rate(i), urate(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2730                  format(1x,i6,4f4.0,2f10.6,10f4.0)
                      write(lunaug,2130) rate(i), urate(i), &
                           d1, d2, d1, d2, &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2
2130                  format(1x,2f10.6,4f4.0,2f6.2,18f4.0)
                   elseif(quadN .eq. 3) then
                      write(lunrh,2840) revl(i),  &
                           d1, d2, d1, d2,  &
                           d1, d2, d1, d2, &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2
2840                  format(1x,i6,8f4.0,2f6.2,6f4.0)
                      write(lunrr,2740) revl(i),  &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2,  &
                           rate(i), urate(i), d1, d2, &
                           d1, d2, d1, d2
2740                  format(1x,i6,8f4.0,2f10.6,6f4.0)
                      write(lunaug,2140) rate(i), urate(i), &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           hard2(i), uhard2(i), d1, d2, &
                           d1, d2, d1, d2
2140                  format(1x,2f10.6,8f4.0,2f6.2,14f4.0)
                   elseif(quadN .eq. 4) then
                      write(lunrh,2850) revl(i),  &
                           d1, d2, d1, d2,  &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           hard2(i), uhard2(i), d1, d2
2850                  format(1x,i6,12f4.0,2f6.2,2f4.0)
                      write(lunrr,2750) revl(i),  &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2,  &
                           d1, d2, d1, d2, &
                           rate(i), urate(i), d1, d2
2750                  format(1x,i6,12f4.0,2f10.6,2f4.0)
                      write(lunaug,2150) rate(i), urate(i), &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           d1, d2, d1, d2, &
                           hard2(i), uhard2(i), d1, d2
2150                  format(1x,2f10.6,12f4.0,2f6.2,10f4.0)
                   endif
                end if
             endif
             !  Check to see whether the observation CCD is in an odd state
             !  and select whether the data will be kept or not

             ikeep = 1

             !  Keep the data set so transfer the information

             if(ikeep .eq. 1) then
                nobs = nobs + 1
                revl(nobs) = revl(i)
                expo(nobs) = expo(i)
                rate(nobs) = rate(i)
                urate(nobs) = urate(i)
                hard2(nobs) = hard2(i)
                uhard2(nobs) = uhard2(i)
                do ii=1,4096
                   specBig(ii,nobs) = specBig(ii,i)
                   specBigOOT(ii,nobs) = specBigOOT(ii,i)
                enddo
             endif
          endif
       end if
    end do
    !
    ! SS: Set the energies of the bins
    !
    allocate(energ(chans))
    do i=1,nChans ! 4096
       energ(i) = 0.005*(i - 1)
    enddo
    !
    ! SS: Get some temporary values for the average rate and hardness
    !     errors for the corner data
    ! BHP:  note: c_tim is NEVER eq 0.0, so that was eliminated
    murate = 0.0
    muhard = 0.0
    do i=1,goodRows ! last-filtered 'best' QPB rows, was nobs
       murate = murate + urate(i) 
       muhard = muhard + uhard2(i) 
    enddo
    if (goodRows .ne. 0) then 
       murate = murate/goodRows
       muhard = muhard/goodRows
    else
       call error("noGoodRows","No rows meet criteria in QPB file!")
    endif
    !
    write(cquad,'(i1)')quadN
    write(errstr,'(A,f12.5,A,f12.5)')'Rate: ',ratein,' Rate error (murate): ',murate
    call message(AppMsg,Sparse)
    write(errstr,'(A,f12.5,A,f12.5)')'Hard: ',hardin,' Hard error (muhard): ',muhard
    call message(AppMsg,Sparse)
    ! good to here!!!
    !  Find the closest c_tim secs
    !
    allocate(dist(goodRows),ddist(goodRows))
    allocate(hdist(goodRows),rdist(goodRows))
    allocate(tdist(goodRows),ord(goodRows))
    ! BHP: removed checks for c_tim/c_dst. They are hardcoded and will
    !      never (unless we WANT them to) fail the if block
    ! SS: First, set the distance in parameter space
    !
    write(errstr,*)'c_dst = ',c_dst
    call message(AppMsg,Noisy)
    do i=1,goodRows
       if(c_dst(1) .gt. 0.0) then
          rdist(i) = (rate(i)-ratein)/c_dst(1) ! ratein=c_val(1)
       else
          rdist(i) = 0.0
       endif
       if(c_dst(2) .ne. 0.0) then
          hdist(i) = (hard2(i)-hardin)/c_dst(2)
       else
          hdist(i) = 0.0
       endif
       if(c_dst(3) .ne. 0.0) then
          ddist(i) = (revl(i)-revlin)/c_dst(3)
       else
          ddist(i) = 0.0
       endif
       dist(i) = sqrt(rdist(i)**2 + hdist(i)**2 + &
            ddist(i)**2)
    enddo
    !
    ! SS: Sort through the CalDB data to order them by the distance
    !     from the observation values.
    ! BHP: Sort through new qdp file in testprods/esas_data for SAS-20+
    !
    do i=1,goodRows
       ord(i) = i
       tdist(i) = dist(i)
    enddo
    do ii=goodRows-1,1,-1
       do i=1,ii
          if(tdist(i) .gt. tdist(i+1)) then
             td = tdist(i)
             tdist(i) = tdist(i+1)
             tdist(i+1) = td
             itd = ord(i)
             ord(i) = ord(i+1)
             ord(i+1) = itd
          endif
       enddo
    enddo

    ! SS:  Now accumulate spectra until a limiting exposure is reached

    nRows1   = 0 ! OLD: n ; easier to grep for in the code
    cum_exp  = 0.0
    ave_rate = 0.0

    do while ((cum_exp .lt. c_tim) .and. (nRows1 .lt. goodRows))
       nRows1 = nRows1 + 1
       cum_exp = cum_exp + expo(ord(nRows1))
       ave_rate = ave_rate + rate(ord(nRows1))*expo(ord(nRows1))
       if(doplot) then
          modd = mod(nRows1,4)
          if (modd .eq. 0) then
             if(quadN .eq. 1) then
                write(lunrh,2825) revl(ord(nRows1)), d1, d2, &
                     hard2(ord(nRows1)), uhard2(ord(nRows1)), d1, &
                     d2, d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
2825            format(1x,i6,2f4.0,2f6.2,12f4.0)
                write(lunrr,2725) revl(ord(nRows1)), d1, d2, &
                     rate(ord(nRows1)), urate(ord(nRows1)), d1, &
                     d2, d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
2725            format(1x,i6,2f4.0,2f10.6,12f4.0)
                write(lunaug,2125) rate(ord(nRows1)),urate(ord(nRows1)), &
                     d1, d2, hard2(ord(nRows1)), uhard2(ord(nRows1)), d1, &
                     d2, d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
2125            format(1x,2f10.6,2f4.0,2f6.2,20f4.0)             
             elseif(quadN .eq. 2) then
                write(lunrh,2835) revl(ord(nRows1)), d1, d2, d1, d2, d1, &
                     d2, hard2(ord(nRows1)), uhard2(ord(nRows1)), d1, d2, d1, &
                     d2, d1, d2, d1, d2
2835            format(1x,i6,6f4.0,2f6.2,8f4.0)
                write(lunrr,2735) revl(ord(nRows1)), d1, d2, d1, d2, d1, d2, &
                     rate(ord(nRows1)), urate(ord(nRows1)), d1, d2, d1, d2, d1, d2, &
                     d1, d2
2735            format(1x,i6,6f4.0,2f10.6,8f4.0)
                write(lunaug,2135) rate(ord(nRows1)),urate(ord(nRows1)), d1, d2, &
                     d1, d2, d1, d2, hard2(ord(nRows1)), uhard2(ord(nRows1)), d1, &
                     d2, d1, d2, d1, d2, d1, d2
2135            format(1x,2f10.6,6f4.0,2f6.2,16f4.0)
             elseif(quadN .eq. 3) then
                write(lunrh,2845) revl(ord(nRows1)), d1, d2, d1, d2, d1, d2, d1, &
                     d2, d1, d2, hard2(ord(nRows1)), uhard2(ord(nRows1)), d1, d2, &
                     d1, d2
2845            format(1x,i6,10f4.0,2f6.2,4f4.0)
                write(lunrr,2745) revl(ord(nRows1)), d1, d2, d1, d2, d1, d2, d1, &
                     d2, d1, d2, rate(ord(nRows1)), urate(ord(nRows1)), d1, d2, &
                     d1, d2
2745            format(1x,i6,10f4.0,2f10.6,4f4.0)
                write(lunaug,2145) rate(ord(nRows1)), urate(ord(nRows1)), d1, d2, d1,&
                     d2, d1, d2, d1, d2, d1, d2, hard2(ord(nRows1)), &
                     uhard2(ord(nRows1)), d1, d2, d1, d2
2145            format(1x,2f10.6,10f4.0,2f6.2,12f4.0)
             elseif(quadN .eq. 4) then
                write(lunrh,2855) revl(ord(nRows1)), d1, d2, d1, d2, d1, d2, d1, &
                     d2, d1, d2, d1, d2, d1, d2, hard2(ord(nRows1)), uhard2(ord(nRows1))
2855            format(1x,i6,14f4.0,2f6.2)
                write(lunrr,2755) revl(ord(nRows1)), d1, d2, d1, d2, d1, d2, d1, &
                     d2, d1, d2, d1, d2, d1, d2, rate(ord(nRows1)), urate(ord(nRows1))
2755            format(1x,i6,14f4.0,2f10.6)
                write(lunaug,2155) rate(ord(nRows1)), urate(ord(nRows1)), d1, d2, d1, &
                     d2, d1, d2, d1, d2, d1, d2, d1, d2, d1, d2, &
                     hard2(ord(nRows1)), uhard2(ord(nRows1))
2155            format(1x,2f10.6,14f4.0,2f6.2,8f4.0)
             endif
          endif
       end if
    enddo
    if (cum_exp .ne. 0) then 
       ave_rate = ave_rate/cum_exp
    else
       call error("expLEzero","Augmented spectrum cumulative exposure LE 0.")
    end if
    write(errstr,4010) 'Exposure                      : ',cum_exp
    call message(AppMsg,Always)
    write(errstr,4010) 'Exposure weighted count rate  : ',ave_rate
    call message(AppMsg,Always)
    write(errstr,4020)  'Observation count rate        : ',ratein, &
         ' and ratio: ', ave_rate/ratein
    call message(AppMsg,Always)
    write(errstr,4010)  'Observation hardness ratio    : ',hardin
    call message(AppMsg,Always)
4010 format(A,f15.6)
4020 format(A,f15.6,A,f15.6)
    !
    ! SS: Actually put together the augmented spectrum augmented exposure
    ! 
    aug_expo = 0.
    aug_spec = 0.
    do i=1,nRows1
       do ii=1,nChans ! 4096
          aug_spec(ii) = aug_spec(ii) + (specBig(ii,ord(i)) - &
               allootsca(ord(i))*specBigOOT(ii,ord(i)))*expo(i)
       enddo
       aug_expo = aug_expo + expo(ord(i))
    enddo
    do i=1,nChans
       aug_spec(i) = aug_spec(i)/aug_expo
    end do
    te1 = 0.0
    te2 = 0.0
    do i=plc_lo(1),plc_lo(2)
       te1 = te1 + aug_spec(i)
    enddo
    do i=plc_hi(1),plc_hi(2)
       te2 = te2 + aug_spec(i) 
    enddo

    write(errstr,4040) te1+te2
4040 format('Augmented count rate    : ',f10.6)
    call message(AppMsg,Always)
    write(errstr,4050) te2/te1
4050 format('Augmented hardness ratio: ',f10.6)
    call message(AppMsg,Always)
    write(errstr,4060) aug_expo
4060 format('Augmented exposure: ',f15.4)
    call message(AppMsg,Always)
    if(doplot) then
       do i=1,nChans
          if(quadN .eq. 1) then
             write(lunspec,2225) energ(i), c_tim*aug_spec(i), &
                  d1, d1, d1 
2225         format(1x,f10.3,f10.5,3f4.0)
          elseif(quadN .eq. 2) then
             write(lunspec,2235) energ(i), d1, c_tim*aug_spec(i), &
                  d1, d1
2235         format(1x,f10.3,f4.0,f10.5,2f4.0)
          elseif(quadN .eq. 3) then
             write(lunspec,2245) energ(i), d1, d1, &
                  c_tim*aug_spec(i), d1
2245         format(1x,f10.3,2f4.0,f10.5,f4.0)
          elseif(quadN .eq. 4) then
             write(lunspec,2255) energ(i), d1, d1, d1, &
                  c_tim*aug_spec(i)
2255         format(1x,f10.3,3f4.0,f10.5)
          endif
       enddo
    endif
    return

  end subroutine augment_pn
  !
  ! old call:
  ! subroutine spscale (xmmdir,xmmd,mosnum,maskfile,specfile,back,c)
  !
  subroutine spscale (mask_F,spec_F,backS,spsum)
    !
    !  Date: 15 November 2006 (SLS)
    !  Conversion to F90 - 7 November 2012 (BHP)
    !  XMM-SAS CALDB compliance - 2021 (BHP)
    !
    !  Calculates the relative contribution of SP flares by region
    !
    !  spscale uses SP flare detector maps and region images to
    !  find the average SP flare contribution over the region.  It
    !  also returns the BACKSCALE value from the spectrum for the
    !  region.
    !
    ! This routine is called by sppartial and protonscale ESAS task
    !
    !  Input:
    !    Mask image file name
    !    Spectrum file name
    !  Output:
    !    BACKSCAL value for the region
    !    Average SP intensity for the region

    type(DataSetT) :: mask_S, spec_S
    type(TableT)   :: mask_T, spec_T
    type(ArrayT)   :: mask_A, spec_A
    character(len=*),intent(in) :: mask_F, spec_F
    character(len=2) :: instM, instS
    character(len=5) :: expidM, expidS
    character(len=6) :: filterM, filterS, type
    character(len=10) :: obsidM, obsidS
    character(len=25) :: smode

    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32, mask_IP
    integer :: XsizeM, YsizeM, i, ii, n, status, typeA
    integer :: typeF, frmtime, revoM, revoS, elow, ehigh, nRows
    logical :: warndie
    real :: backM, backS, areaM, areaS, expoM, expoS, spsum
    real(kind=single), dimension(:,:), pointer :: aDR32, spdetmap_RP
    real(kind=double), dimension(:,:), pointer :: aDR64                                       
    !
    ! Read in the mask image (e.g.: mos2S002-maskim-400-750.fits)
    ! Mask file is a 780x780 I4 primary image array
    !
    write(errstr,*)"Image file    : ",trim(adjustl(mask_F))
    call message(AppMsg,Always)
    write(errstr,*)"Spectral file : ",trim(adjustl(spec_F))
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(adjustl(mask_F)),mask_S,1,status,warndie)
    warndie = .TRUE.     ! see read_header_info in gofio_mod for behavior                
    typeF = 3 ! 1 = evli, 2 = spec, 3 = image                                            
    call read_header_info(typeF,mask_S,mask_T,mask_A,instM,obsidM,&            
         expidM,smode,frmtime,filterM,revoM,backM,areaM,expoM,elow,&               
         ehigh,nRows,warndie)
    backM = backM / 1440000.0
    filterM = to_upper(filterM)
    call read_2Darray(mask_S,mask_A,aDR32,aDR64,aDI16,mask_IP,&              
         XsizeM,YsizeM,typeA)
    !
    ! Read in spectral file
    !
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(adjustl(spec_F)),spec_S,1,status,warndie)
    warndie = .TRUE.     ! see read_header_info in gofio_mod for behavior                
    typeF = 2 ! 1 = evli, 2 = spec, 3 = image                                            
    call read_header_info(typeF,spec_S,spec_T,spec_A,instS,obsidS,&             
         expidS,smode,frmtime,filterS,revoS,backS,areaS,expoS,elow,&               
         ehigh,nRows,warndie)
    filterS = to_upper(filterS)        
    backS = backS / 1440000.0
    ! sanity checks
    if ((XsizeM .ne. 780) .or. (YsizeM .ne. 780)) then
       write(errstr,*)"spscale: image not 780x780! X/Y = ",XsizeM,YsizeM
       call warning("not780x780",trim(errstr))
    endif
    if (trim(adjustl(obsidM)) .ne. trim(adjustl(obsidS))) then 
       call error("notSAMEobsid","Image and spectrum have different obsids.")
    end if
    if (trim(adjustl(expidM)) .ne. trim(adjustl(expidS))) then 
       call warning("notSAMEexpid","Image and spectrum have different expids.")
    end if
    if (revoM .ne. revoS) then
       call error("notSAMEexpid","Image and spectrum have different expids.")
    end if
    if (trim(adjustl(filterM)) .ne. trim(adjustl(filterS))) then 
       call warning("notSAMEfilter","Image and spectrum have different filters.")
    end if
    ! diagnostics
    write(errstr,*)type," inst mask    = ",instM
    call message(AppMsg,Verbose)
    write(errstr,*)type," inst spec    = ",instS
    call message(AppMsg,Verbose)
    write(errstr,*)type," obsid mask   = ",obsidM
    call message(AppMsg,Verbose)
    write(errstr,*)type," obsid spec   = ",obsidS
    call message(AppMsg,Verbose)
    write(errstr,*)type," expid mask   = ",expidM
    call message(AppMsg,Verbose)
    write(errstr,*)type," expid spec   = ",expidS
    call message(AppMsg,Verbose)
    write(errstr,*)type," revolut mask = ",revoM
    call message(AppMsg,Verbose)
    write(errstr,*)type," revolut spec = ",revoS
    call message(AppMsg,Verbose)
    write(errstr,*)type," filter mask  = ",filterM
    call message(AppMsg,Verbose)
    write(errstr,*)type," filter spec  = ",filterS
    call message(AppMsg,Verbose)
    !
    ! get the SPDETMAP from the CAL
    !
    write(errstr,*)"Calling CAL_setState..."
    call message(AppMsg,Verbose)
    call CAL_setState(mask_F)
    write(errstr,*)"Calling CAL_spdetmap..."
    call message(AppMsg,Verbose)
    call CAL_spdetmap(spdetmap_RP)
    !
    !  Loop over the image
    !
    write(errstr,*)'startLoop = ',XsizeM,YsizeM
    call message(AppMsg,Verbose)
    n = 0
    spsum = 0.
    do i=0,XsizeM-1
       do ii=0,YsizeM-1
          if (mask_IP(i,ii) .gt. 0) then
             n = n + 1
             spsum = spsum + spdetmap_RP(i,ii)
          endif
       enddo
    enddo
    spsum = spsum / real(n)
    ! moved here to 'fix' seg fault on linux, 12/2023 BHP
    write(errstr,*)'Freeing up mask and spec memory'
    call message(AppMsg,Verbose)
    call release(mask_S)
    call release(spec_S)
    return

  end subroutine spscale
  !
  !
  !
  subroutine hardscale (instrume,filter,alpha,low,high,scalefactor)
    !
    !
    ! called by combimage and merge_comp_xmm
    !
    ! instrume = EMOS1, EMOS2, EPN
    ! filter = THIN1, MEDIUM, THICK
    ! alpha = 1.0 < alpha < 2.4 photon index
    !
    character(len=8) :: instrume
    character(len=6) :: filter
    character(len=6) :: idxband
    !
    integer :: x = 240, y = 240 ! adjust here later???!!!
    integer :: i, j, foo
    integer, intent(in) :: low, high
    !
    real(kind=single), intent(in) :: alpha
    real(kind=single), intent(out) :: scalefactor
    real(kind=single) :: ts, th
    real(kind=single), dimension(:,:), pointer :: sfmaps_RP, sfmapm_RP, &
         sfmaph_RP, sfmapt_RP
    !
    write(errstr,*)"Calling CAL_setState with inst=",trim(adjustl(instrume)),&
         ' filter = ',trim(adjustl(filter))
    call message(AppMsg,Noisy)
    call CAL_setState(instrument=instId(trim(instrume)),&
         filterId=filterIdToInt(trim(filter)))
    !
    ! calculate combined scalefactor array
    ! BHP: fix this for roundoff???!!
    if(alpha .eq. 1.0) then
       idxband = 'Hard'
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmapt_RP) ! only hard
    elseif(alpha .eq. 1.7) then
       idxband = 'Medium'
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmapt_RP) ! only medium
    elseif(alpha .eq. 2.4) then
       idxband = 'Soft'
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmapt_RP) ! only soft
    elseif(alpha .lt. 1.7) then ! between medium and hard
       idxband = "Hard"
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmaph_RP)
       idxband = "Medium"
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmapm_RP)
       ts = (1.7 - alpha)/0.7
       th = (alpha - 1.0)/0.7
       do i = 0, x-1 ! 240, pointer arrays from 0!
          do j = 0, y-1 ! 240
             sfmapt_RP(i,j) = (ts * sfmapm_RP(i,j)) + &
                  (th * sfmaph_RP(i,j))
          enddo
       enddo
    else
       idxband = "Soft"
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmaps_RP)
       idxband = "Medium"
       write(errstr,*)'alpha = ',alpha,' idxband = ',idxband
       call message(AppMsg,Noisy)
       call CAL_scalefactormap(idxband, sfmapm_RP)
       ts = (2.4 - alpha)/0.7
       th = (alpha - 1.7)/0.7
       do i = 0, x-1 ! 240
          do j = 0, y-1 ! 240
             sfmapt_RP(i,j) = (ts * sfmaps_RP(i,j)) + &
                  (th * sfmapm_RP(i,j))
          enddo
       enddo
    endif
    !
    !  Now evaluate the scale factor. low high are calculated in combimage as:
    !  low = (low + 25)/50, high = (high + 25)/50
    !
    write(errstr,*)'Calculating scalefactor...'
    call message(AppMsg,Noisy)
    !
    scalefactor = sfmapt_RP(low-1,high-1) ! pointers go from 0!!!
    if ((idxband .eq. "Medium") .and. (index(instrume,'2') .ne. 0)) scalefactor = 1.
    write(errstr,*)'Scalefactor = ',scalefactor
    call message(AppMsg,Noisy) 
    !
    return

  end subroutine hardscale

end module esas_mod
