! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
module jplephemeris_mod

  use dal
  use types
  use param 
  use errorHandling
  use gofconstants_mod
  use gofutils_mod

  implicit none 

  public

  TYPE :: jplephem
     character(len=1024) :: filename
     integer(kind=Int32) :: edited
     character(len=30) :: creation_date
     character(len=80) :: author
     integer :: nrows
     real(kind=double) :: tstart
     real(kind=double) :: tstop
     real(kind=double) :: timedel
     character(len=11) :: format
     integer(kind=Int32) :: denum
     real(kind=double) :: c            ! speed of light
     real(kind=double) :: emrat
     real(kind=double) :: au           ! astronomical unit
     real(kind=double) :: msol         ! solar mass
     real(kind=double) :: sunrad
     real(kind=double) :: jdlimits(2)
     integer :: jdrows
     character(len=22) :: objname(13)
     integer(kind=Int16) :: ptr(13)
     integer(kind=Int16) :: ncoeff(13)
     integer(kind=Int16) :: nsub(13)
     character(len=6) :: names(156)
     real(kind=double) :: values(156)
  END TYPE jplephem

contains

  subroutine jplephread (jplephF,info,raw,jdlimits)

    type(DataSetT) :: jplephS
    type(TableT) :: de1T, de2T, de3T
    type(ColumnT) :: cnameC, cvalueC, objectC, pointerC, numcoeffC, &
         numsubintvC, chebcoeffsC, tstartC, tstopC
    type (jplephem) :: info
!
    real(kind=double), dimension(:,:), pointer :: raw
    real(kind=double), dimension(:), pointer :: x_ptr, y_ptr, z_ptr, &
         gx_ptr, gy_ptr, gz_ptr
    character(len=1024) :: jplephF  ! jpl ephemeris file, JPLEPH.405 for now
    character(len=6), dimension(:), pointer :: cnames_ptr
    character(len=22), dimension(:), pointer :: object_ptr
    integer(kind=Int16),  dimension(:), pointer :: pointer_ptr, numcoeff_ptr,&
         numsubintv_ptr
    integer :: nrows1, nrows2, nrows3, i, rowlimits(2), ntimes, dims, &
         ff, j, k
    logical :: fatal
    real(kind=double) :: jdlimits(2)
    real(kind=double) :: sun_ra, sun_dec, sun_elon, sun_elat, sun_r
    real(kind=double) :: sun_x, sun_y, sun_z
    real(kind=double) :: au, clight, msol, radsol, default, xx
    real(kind=single) :: timedel, tstart, tstop
    real(kind=double), dimension(:), pointer :: jdtimes
    real(kind=double), dimension(:), pointer :: cvalues_ptr, &
         tstart_ptr, tstop_ptr
    real(kind=double), pointer :: chebcoeffs_ptr(:,:) 

! first, open and read JPLEPH.405 file
    write(errstr,*) "Reading JPLEPH.405"
    call message(AppMsg,Sparse)

    jplephS = dataSet(trim(jplephF), READ)

! first extension should always be DE1 Planetary Ephemeris DE405/LE405

    write(errstr,*) "Checking each block in JPLEPH.405 for consistency"
    call message(AppMsg,Sparse)

    if (.not.hasBlock(jplephS,"DE1")) then
       call error("badJPLephFile1",&
            "JPL ephemeris file does not contain a XXXX extension")
    else
       de1T = table(jplephS,"DE1")
       nrows1 = numberOfRows(de1T)
    endif
    write(errstr,*) "Allocating ",nrows1," for 1st ephem table"
    call message(AppMsg,Sparse)

    allocate(cnames_ptr(nrows1))
    allocate(cvalues_ptr(nrows1))

! read the two columns, Cname and Cvalue from table DE1

    if (.not.hasColumn(de1T,"Cname")) then
       call error("badTable","DE1 table does not contain a Cname column")
    else
       cnameC = column(de1T,"Cname",READ)
       do i = 1,nrows1
          cnames_ptr(i) = trim(stringCell(cnameC,i-1))
        end do
    endif
     if (.not.hasColumn(de1T,"Cvalue")) then
       call error("badTable","DE1 table does not contain a Cvalue column")
    else
       cvalueC = column(de1T,"Cvalue",READ)
       cvalues_ptr => real64Data(cvalueC)
    endif

! second extension should always be DE2 Chebycheff coefficients DE405/LE405

    if (.not.hasBlock(jplephS,"DE2")) then
       call error("badJPLephFile2",&
            "JPL ephemeris file does not contain a XXXX extension")
    else
       de2T = table(jplephS,"DE2")
       nrows2 = numberOfRows(de2T)
    endif
    write(errstr,*) "Allocating ",nrows2," for 2nd ephem table"
    call message(AppMsg,Sparse)

    allocate(object_ptr(nrows2))
    allocate(pointer_ptr(nrows2))
    allocate(numsubintv_ptr(nrows2))
    allocate(numcoeff_ptr(nrows2))

! read the two columns, Cname and Cvalue from table DE1

    if (.not.hasColumn(de2T,"Object")) then
       call error("badTable2","DE2 table does not contain an Object column")
    else
       objectC = column(de2T,"Object",READ)
       do i = 1,nrows2
          object_ptr(i) = trim(stringCell(objectC,i-1))
       end do
    endif
    if (.not.hasColumn(de2T,"Pointer")) then
       call error("badTable2","DE2 table does not contain a Pointer column")
    else
       pointerC = column(de2T,"Pointer",READ)
       pointer_ptr => int16Data(pointerC)
    endif
    if (.not.hasColumn(de2T,"NumCoeff")) then
       call error("badTable2","DE2 table does not contain a NumCoeff column")
    else
       numcoeffC = column(de2T,"NumCoeff",READ)
       numcoeff_ptr => int16Data(numcoeffC)
    endif
    if (.not.hasColumn(de2T,"NumSubIntv")) then
       call error("badTable2","DE2 table does not contain a NumSubIntv column")
    else
       numsubintvC = column(de2T,"NumSubIntv",READ)
       numsubintv_ptr => int16Data(numsubintvC)
    endif
!**** need this in f95**** if necessary???...
!      ;; Trim each object name to first word only
!      for i = 0, n_elements(ephobj)-1 do begin
!          ephobj(i) = strupcase((str_sep(ephobj(i), ' '))(0))
!      endfor

!
! third (chebycheff coefficients extension)
!
    if (.not.hasBlock(jplephS,"DE3")) then
       call error("badJPLephFile3",&
            "JPL ephemeris file does not contain a DE3 extension")
    else
       de3T = table(jplephS,"DE3")
       nrows3 = numberOfRows(de3T)            ! keep this nrows for later
    endif
    write(errstr,*) "Allocating ",nrows3," for 3rd ephem table"
    call message(AppMsg,Sparse)
    allocate(chebcoeffs_ptr(nrows3,1018))

    if (.not.hasColumn(de3T,"ChebyCoeffs")) then
       call error("badTable3","DE3 table does not contain a ChebCoeffs column")
    else
       chebcoeffsC = column(de3T,"ChebCoeffs",READ)
       chebcoeffs_ptr => real64VectorData(chebcoeffsC)
    endif
    if (.not.hasColumn(de3T,"start")) then
       call error("badTable3","DE3 table does not contain a start column")
    else
       tstartC = column(de3T,"start",READ)
       tstart_ptr => real64Data(tstartC)
    endif
    if (.not.hasColumn(de3T,"stop")) then
       call error("badTable3","DE3 table does not contain a stop column")
    else
       tstopC = column(de3T,"stop",READ)
       tstop_ptr => real64Data(tstopC)
    endif
!
! get some keywords, I mean, attributes
!
    if (hasAttribute(de3T,'TIMEDEL')) then
       timedel = real32Attribute(de3T,'TIMEDEL')
   else
       call error("noJPLTIMEDEL","TIMEDEL attribute missing from ephem file.")
    end if
    if (hasAttribute(de3T,'TSTART')) then
       tstart = real32Attribute(de3T,'TSTART')
    else
       call error("noJPLTSTART","TSTART attribute missing from ephem file.")
    end if
    if (hasAttribute(de3T,'TSTOP')) then
       tstop = real32Attribute(de3T,'TSTOP')
    else
       call error("noJPLTSTOP","TSTOP attribute missing from ephem file.")
    end if
!
    if (floor((tstop-tstart + 0.5)/timedel) .ne. nrows3) then 
       call error("floorWrong","Incorrect number of rows in ephem file")
    endif
!
! always call with jdlimits(2), even if 0,0 (default for all elements, 
! defaulting to tstart, tstop???

    if (jdlimits(1) .lt. tstart .or. jdlimits(2) .gt. tstop) then
       call error("ERROR: ephem file does not cover the time of interest")
    else
       rowlimits(1) = floor((jdlimits(1)-tstart)/timedel) - 2
       rowlimits(2) = floor((jdlimits(2)-tstart)/timedel) + 2
       if (rowlimits(1) .lt. 1) then
          rowlimits(1) = 1
          jdlimits(1) = tstart
       end if
       if (rowlimits(2) .gt. nrows3) then
          rowlimits(2) = nrows3
          jdlimits(2) = tstop
       end if
    end if
    write(errstr,*) ' jdlimits(1) = ',jdlimits(1),' jdlimits(2) = ',&
         jdlimits(2)
    call message(AppMsg,Always)
    write(errstr,*)'jdlimits(2) - jdlimits(1) =',jdlimits(2) - jdlimits(1)
    call message(AppMsg,Always)
    write(errstr,*)' tstart = ',tstart,'  tstop = ',tstop
    call message(AppMsg,Always)
    print*,' rowlimits(1,2) = ',rowlimits(1),' ', rowlimits(2)
    dims = size(chebcoeffs_ptr,1)
    write(errstr,*)'dims = ',dims
    info%jdlimits(1) = jdlimits(1)
    info%jdlimits(2) = jdlimits(2)
!
! now, have all data, fill the "info" structure
!
    default = -1d0
    fatal = .true.
    info%filename = trim(jplephF)
    info%edited = 0
    info%creation_date = ''
    info%author = ''
    info%nrows =  nrows3 ! hard coded into ephem file
    info%jdrows = rowlimits(2) - rowlimits(1) + 1
    info%tstart = 2433264.5 ! ephem file range 0 = 12/14/1949:12:00:00
    info%tstop = 2469808.5  ! ephem file range 1 = 01/01/2050:12:00:00
    info%timedel = timedel
    info%format = 'BINEPH2FITS'
    info%denum = nint(jplephemval(default,fatal,'DENUM',nrows1,cnames_ptr,cvalues_ptr))    
    clight = jplephemval(default,fatal,'CLIGHT',nrows1,cnames_ptr,cvalues_ptr)
    if (clight .eq. 0) then
       call warning("noClight","could not load clight value from DE2")
       clight = 2.9979246e+05 ! km/s, should never happen unless JPL corrupt
    end if
    info%emrat = 1.0d0 / (1.0d0 + &
         jplephemval(default,fatal,'EMRAT',nrows1,cnames_ptr,cvalues_ptr))
    au = jplephemval(default,fatal,'AU',nrows1,cnames_ptr,cvalues_ptr)
    msol =  jplephemval(default,fatal,'GMS',nrows1,cnames_ptr,cvalues_ptr)
    radsol = jplephemval(default,fatal,'ASUN',nrows1,cnames_ptr,cvalues_ptr)
    write(errstr,*)'radsol = ',radsol
    call message(AppMsg,Sparse)
    xx = au / clight ! au in light sec
    write(errstr,*)'xx = ',xx,' msol = ',msol
    call message(AppMsg,Sparse)
    msol = msol * xx * xx * xx / 86400.0d0**2 ! GM_sun in lt sec
    write(errstr,*)'msol now = ',msol
    call message(AppMsg,Sparse)
    radsol = radsol / clight    
    clight = clight * 1000d0 ! speed of light in m/s
    info%au = 1000d0 * au / clight
    info%c = clight
    info%msol = msol
    info%sunrad = radsol
!
! set jdlimits, rowlimits
!
    info%objname(1:nrows2) = object_ptr(1:nrows2)
    info%ptr(1:nrows2) = pointer_ptr(0:nrows2-1)
    info%ncoeff(1:nrows2)  = numcoeff_ptr(0:nrows2-1) 
    info%nsub(1:nrows2)  = numsubintv_ptr(0:nrows2-1)
    info%names(1:nrows1) = cnames_ptr(1:nrows1)
    info%values(1:nrows1) = cvalues_ptr(0:nrows1-1) 

    write(errstr,*)'rowlimits = ',rowlimits(1),' ',rowlimits(2)
    call message(AppMsg,Sparse)
    write(errstr,*)'filename = ',trim(info%filename)
    call message(AppMsg,Sparse)
    write(errstr,*)'edited = ',info%edited
    call message(AppMsg,Sparse)
    write(errstr,*)'creation = ',trim(info%creation_date)
    call message(AppMsg,Sparse)
    write(errstr,*)'auth = ',trim(info%author)
    call message(AppMsg,Sparse)
    write(errstr,*)'nrows = ',info%nrows
    call message(AppMsg,Sparse)
    write(errstr,*)'tstart = ',info%tstart
    call message(AppMsg,Sparse)
    write(errstr,*)'tstop = ',info%tstop
    call message(AppMsg,Sparse)
    write(errstr,*)'timedel = ',info%timedel
    call message(AppMsg,Sparse)
    write(errstr,*)'denum = ',info%denum
    call message(AppMsg,Sparse)
    write(errstr,*)'c = ',info%c
    call message(AppMsg,Sparse)
    write(errstr,*)'emrat = ',info%emrat
    call message(AppMsg,Sparse)
    write(errstr,*)'au = ',info%au
    call message(AppMsg,Sparse)
    write(errstr,*)'msol = ',info%msol
    call message(AppMsg,Sparse)
    write(errstr,*)'sunrad = ',info%sunrad
    call message(AppMsg,Sparse)
    write(errstr,*)'jdlimits = ',info%jdlimits
    call message(AppMsg,Sparse)
    write(errstr,*)'jdrows = ',info%jdrows
    call message(AppMsg,Sparse)
    write(errstr,*)'objname = ',info%objname
    call message(AppMsg,Sparse)
    
! fill the raw array
!
    
    rowlimits(2) = 576
    rowlimits(1) = 572

    write(errstr,*)'allocating raw array with dims, rowlimits =',&
         dims,rowlimits(2)-rowlimits(1)+1
    call message(AppMsg,Sparse)
    
    allocate (raw(dims, rowlimits(2)-rowlimits(1)+1))
!    raw = reshape(chebcoeffs_ptr, &
!         (/ dims, rowlimits(2)-rowlimits(1)+1 /))
    k = 0
    do i=1,1018
       do j=rowlimits(1),rowlimits(2)
          k = k + 1
          raw(i,k) = chebcoeffs_ptr(i,j)
       end do
    end do
!    raw = chebcoeffs_ptr(0:1017,rowlimits(1):rowlimits(2))
    do i=0,1017
       do j=0,4
          ff = floor(raw(i,j))
          write(errstr,*)i,j,ff,"***"
          call message(AppMsg,Noisy)
       end do
    end do
    stop
    write(errstr,*)'raw(coeffs',',',dims,',',(rowlimits(2)-rowlimits(1)+1),')'
    call message(AppMsg,Noisy)
    write(errstr,*)'raw=',shape(raw)
    call message(AppMsg,Noisy)
    do i = 1,10
       write(errstr,*)'i = ',i,' raw = ',raw(i,1)       
       call message(AppMsg,Noisy)
    end do
    stop
!    time = 2450814.0d0 + (time / 86400.0d0)
!    call jplephinterp(info,chebcoeffs_ptr,time_ptr,sun_x,sun_y,sun_z)
!    print *,'x,y,z=',sun_x,' ',sun_y,' ',sun_z
!@    sun_r = sqrt( sun_x**2 + sun_y**2 + sun_z**2 )
!   sun_ra = datan2(sun_y, sun_x) * 180. / Pi
!    if (sun_ra .lt. 0.) then 
!       sun_ra = sun_ra + 360.d0
!    end if
!    sun_dec = asin(sun_z/sun_r) * 180. / Pi
!    write(errstr,*)'sun position in celestial: ',sun_ra,sun_dec

! convert to Ecliptic with euler convert (old euler.pro)

!    call eulerconvert (sun_ra,sun_dec,sun_elon,sun_elat)
!    write(errstr,*)'sun position in ecliptic:',sun_elon,sun_elat

!    deallocate (cvalues_ptr)
!    deallocate (cnames_ptr)
!    deallocate (object_ptr)
!    deallocate (numcoeff_ptr)
!    deallocate (numsubintv_ptr)
!    deallocate (pointer_ptr)
!    deallocate (chebcoeffs_ptr)
!    deallocate (tstart_ptr)
!    deallocate (tstop_ptr)
    return

  end subroutine jplephread
!
!
!
  subroutine jplephinterp(info,raw,time_ptr,x,y,z)
!
! don't worry about objectname=11 ('SUN') or center=earth, they are constant here (for now)
!
    integer :: nelems, i, nc, ns, nr, nseg, ieph, ii1, ii2
    real(kind=double), dimension(:,:), pointer :: raw
    real(kind=double), dimension(:), pointer :: jdtimes, time_ptr 
    real(kind=double) :: sun_x, sun_y, sun_z
! sun_ra, sun_dec, sun_r, sun_elon, sun_elat
    real(kind=double) :: dt, jd0, jd1, tint, time
    real(kind=double) :: p0, p1, v0, v1, v2, tt, x, y, z 
    integer(kind=Int32) :: tseg, i0, i1, i2, ii, jj, j, k, l, ff
    real(kind=double), allocatable, dimension(:,:,:,:) :: coeffs_in,coeffs_out
    type (jplephem) :: info
     
    nc = info%ncoeff(11) ! always 11 for Sun
    ns = info%nsub(11)   ! n subintervals
    dt = info%timedel    ! 
    nr = info%jdrows
!
! tbase is 0 here, unlike in idl
!
    jd0 = info%jdlimits(1) ! full range here
    jd1 = info%jdlimits(2) ! of entire ephemeris

    write(errstr,*) 'nc = ',nc,' ns = ',ns
    write(errstr,*) 'dt = ',dt,' nr = ',nr
    write(errstr,*) 'jd0 = ',jd0,' jd1 = ',jd1

!  Extract coefficient data from RAW
!  if obj EQ 11 then begin (always 11, Sun here, for now???)
!  Nutations have two components for SUN

    ii1 = info%ptr(11) - 1
    write(errstr,*)'infoptr=',info%ptr(11)
    ii2 = ii1 + nc * ns * 3 - 1
    write(errstr,*)'ii1, ii2= ',ii1,' ',ii2
    write(errstr,*)'allocating coeffs array with shape=',nc,3,ns,nr
    allocate(coeffs_in(nc,3,ns,nr))
!      ;; All other bodies are done with three components
      ii1 = info%ptr(10)-1
      ii2 = ii1 + nc*ns*3 - 1
!      coeffs = reform(raw(ii1:ii2,*),nc,3,ns,nr, /overwrite)
!      print,'Nutations have three components (obj=',obj,')'

    coeffs_in = reshape(raw(ii1:ii2,1:5),(/ nc, 3, ns, nr /))
    ii1 = 752
    ii2 = 817
    write(errstr,*)' doing silly loop'
!    do i=ii1,ii2
!       do j=1,5
!          write(errstr,*)'i=',i,' j=',j,' raw(i,j)=',raw(i,j)
!       end do
!    end do
    stop

    do i = 1, nc
       do j = 1, 3
          do k = 1, ns
             do l = 1, nr
                write(errstr,*)'i=',i,'j=',j,'k=',k,'l=',l,&
                     'coeff=',coeffs_in(i,j,k,l)
             end do
          end do
       end do
    end do
    stop

!***    tint = (time - jd0) / dt ! Interval number (real)
    write(errstr,*)'tintt = ',tint,' time = ',time
    ieph = floor(tint)       ! Interval number (index = int)
    tint = (tint-ieph)*ns    ! Subinterval number (real)
    nseg = floor(tint)       ! Subinterval number (index = int)
    print *,'ieph = ',ieph,' tint = ',tint,' nseg = ',nseg

!      coeffs = reform(dblarr(nc,3,ns,nr), nc,3,ns,nr, /overwrite)
!      coeffs[0,0,0,0] = reform(raw[ii1:ii2,*],nc,2,ns,nr, /overwrite)
!  ;; Chebyshev "x" (rescaled to range = [-1,1] over subinterval)

    tseg = 2.0d0 * (tint - nseg) - 1

!  ;; Below is an optimization.  If the time interval doesn't span an
!  ;; ephemeris subinterval, then we can index the coefficient array by
!  ;; a scalar, which is much faster.  Otherwise we maintain the full
!  ;; vector-level indexing.
!    mini(1) = minmax(ieph) 
!    minn = minmax(nseg)
!    if (mini(1) .eq. mini(2) .and. minn(1) .eq. minn(2)) then
!       ieph = ieph(1)
!       nseg = nseg(1)
!    endif
    
!  ;; Initialize the first two Chebyshev polynomials, which are P_0 = 1
!  ;; and P_1(x) = x
    p0 = 1.0d0
    p1 = tseg
!  ;; Initial polynomials for Chebyshev derivatives, V_0 = 0, V_1(x) =
!  ;; 1, V_2(x) = 4*x
    v0 = 0.0d0
    v1 = 1.0d0
    v2 = 4.0d0 * tseg
    tt = 2.0d0 * tseg
    x  = 0.0d0 
    y  = 0.0d0 
    z  = 0.0d0
    i0 = ieph * 1 
    i1 = i0 + 1 
    i2 = i1 + 1
!
!  Compute Chebyshev functions two at a time for efficiency
!  for i = 0, nc-1, 2 do begin
    do i = 1, nc, 2
       if (i .eq. (nc-1)) then
          p1 = 0
          v1 = 0
       endif
       ii = i0 + i
!       jj = i0 + ((i+1) < (nc-1))
       jj = i0 + (min((i+1),(nc-1)))
      write(errstr,*)'ii = ',ii

      x = x + coeffs_in(ii,i0,nseg,ieph)*p0 + coeffs_in(jj,i0,nseg,ieph)*p1
      y = y + coeffs_in(ii,i1,nseg,ieph)*p0 + coeffs_in(jj,i1,nseg,ieph)*p1
      z = z + coeffs_in(ii,i2,nseg,ieph)*p0 + coeffs_in(jj,i2,nseg,ieph)*p1

!          ;; Advance to the next set of Chebyshev polynomials.  For no
!          ;; velocity, we can re-use old variables.
!       p0 = tt * p1 - temporary(p0)
       p0 = tt * p1 - p0
!       p1 = tt*p0 - temporary(p1)
       p1 = tt * p0 - p1
    end do

    return

  end subroutine jplephinterp
  
  subroutine eulerconvert(ai,bi,ao,bo)
!
! adapted from euler.pro IDL program, BHP, 12/09
!
!  J2000 coordinate conversions are based on the following constants
!  (see the Hipparcos explanatory supplement).
!  eps = 23.4392911111d              Obliquity of the ecliptic
!  alphaG = 192.85948d               Right Ascension of Galactic North Pole
!  deltaG = 27.12825d                Declination of Galactic North Pole
!  lomega = 32.93192d                Galactic longitude of celestial equator  
!  alphaE = 180.02322d              Ecliptic longitude of Galactic North Pole
!  deltaE = 29.811438523d            Ecliptic latitude of Galactic North Pole
!  Eomega  = 6.3839743d              Galactic longitude of ecliptic equator             
!
! just need to convert ra-dec to ecliptic (select=3 in euler.pro)
!
    integer :: i
    real(kind=double), intent(in) :: ai, bi
    real(kind=double), intent(out) ::  ao, bo 
    real(kind=double) :: psi(6), stheta(6), ctheta(6), phi(6)
    real(kind=double) :: sb, cb, cbsa
    real(kind=double) :: temp
!
! always J2000, always degrees
!
    psi = (/0.57477043300d0, 4.9368292465d0,  &
         0.00000000000d0, 0.0000000000d0,  &  
         0.11142137093d0, 4.71279419371d0/)     
    stheta =(/0.88998808748d0,-0.88998808748d0, &
         0.39777715593d0,-0.39777715593d0, &
         0.86766622025d0,-0.86766622025d0/)    
    ctheta =(/0.45598377618d0, 0.45598377618d0, &
         0.91748206207d0, 0.91748206207d0, &
         0.49714719172d0, 0.49714719172d0/)    
    phi  = (/4.9368292465d0,  0.57477043300d0, &
         0.0000000000d0, 0.00000000000d0, &
         4.71279419371d0, 0.11142137093d0/)
    
    ao  = ai / Rad2Deg - phi(3)
    bo = bi / Rad2Deg
    sb = sin(bo)
    cb = cos(bo)
    cbsa = cb * sin(ao)
    bo = -stheta(3) * cbsa + ctheta(3) * sb
    if (bo .gt. 1.d0) then
       if ((bo - 1.d0) .gt. 0.00001d0) then
          call error("AsinOutOfRange", &
               "EULERCONVERT: asin(val>1.0)")
       else
! occasional rounding error in 
! bo = -stheta(3) * cbsa + ctheta(3) * sb
          bo = 1.0d0 

       end if
    end if
    bo = asin(bo)
    bo = bo * Rad2Deg
    !
    !       temp = (cb(i) * cos(ao(i))) / &
    !            (ctheta(3) * cbsa(i) + stheta(3) * sb(i))
    ao =  &
         datan2( (ctheta(3) * cbsa + stheta(3) * sb), (cb * cos(ao)) )
    !       ao(i) = ( (ao+psi(3)+d4pi) mod d2pi) 
    ao = mod((ao + psi(3) + d4pi),d2pi)
    ao = ao * Rad2Deg
    return
    
  end subroutine eulerconvert
!
!
! *** function section ***
!
  function jplephemval (default, fatal, name, nrows, names, values)

    ! function to cross-reference a Cvalue from a Cname

    real(kind=double) :: jplephemval
    real(kind=double) :: default
    real(kind=double), dimension(:), pointer :: values
    character(len=*), dimension(:), pointer :: names
    character(len=*) :: name
    integer :: i = 0
    integer, intent(in) :: nrows
    logical :: fatal

    do i = 1, nrows          
       if (trim(names(i)) .eq. trim(name)) then
          jplephemval = values(i-1) ! pointer indexed to 0
          return
       end if
!       if (fatal) call error("JPLEPHEMVAL: unable to parse RADSOL or ASUN")
    end do
    jplephemval = default
    
  end function jplephemval

  subroutine foo

!    real(kind=double), allocatable, dimension(:) :: sun_r, sun_ra, sun_x, sun_y, sun_z
!    real(kind=double), allocatable, dimension(:) :: plc
!    real(kind=double) :: sun_r(1000), sun_ra(1000), sun_x(1000), sun_y(1000), sun_z(1000)
!    real(kind=double) :: plc!

!    sun_r = sqrt(sun_x**2 + sun_y**2 + sun_z**2)
!    sun_ra = atan(sun_x,sun_y) * 180./Pi
!    where(sun_ra .lt. 0.) plc = sun_ra


  end subroutine foo


!+++++++++++++++++++++++++++++++++
!
  subroutine interp(buf,t,ncf,ncm,na,ifl,pv)
!
!+++++++++++++++++++++++++++++++++
!
!     THIS SUBROUTINE DIFFERENTIATES AND INTERPOLATES A
!     SET OF CHEBYSHEV COEFFICIENTS TO GIVE POSITION AND VELOCITY
!
!     CALLING SEQUENCE PARAMETERS:
!
!       INPUT:
!
!         BUF   1ST LOCATION OF ARRAY OF D.P. CHEBYSHEV COEFFICIENTS OF POSITION
!
!           T   T(1) IS DP FRACTIONAL TIME IN INTERVAL COVERED BY
!               COEFFICIENTS AT WHICH INTERPOLATION IS WANTED
!               (0 .LE. T(1) .LE. 1).  T(2) IS DP LENGTH OF WHOLE
!               INTERVAL IN INPUT TIME UNITS.
!
!         NCF   # OF COEFFICIENTS PER COMPONENT
!
!         NCM   # OF COMPONENTS PER SET OF COEFFICIENTS
!
!          NA   # OF SETS OF COEFFICIENTS IN FULL ARRAY
!               (I.E., # OF SUB-INTERVALS IN FULL INTERVAL)
!
!          IFL  INTEGER FLAG: =1 FOR POSITIONS ONLY
!                             =2 FOR POS AND VEL
!
!
!       OUTPUT:
!
!         PV   INTERPOLATED QUANTITIES REQUESTED.  DIMENSION
!               EXPECTED IS PV(NCM,IFL), DP.
!
!
!      implicit double precision (a-h,o-z)
!
    save
!
    integer, intent(in) :: ncf, ncm, na, ifl
    integer :: i, j, l, np, nv
    integer(kind=int32) :: dt1
    real(kind=double) :: buf(ncf,ncm,*),t(2),pv(ncm,*),pc(18),vc(18)
    real(kind=double) :: dna, temp, tc, twot, vfac
!
    data np/2/
    data nv/3/
    data twot/0.d0/
    data pc(1),pc(2)/1.d0,0.d0/
    data vc(2)/1.d0/
!
!       entry point. get correct sub-interval number for this set
!       of coefficients and then get normalized chebyshev time
!       within that subinterval.
!
    dna=dble(na)
    dt1=dint(t(1))
    temp=dna*t(1)
    l=idint(temp-dt1)+1

!         tc is the normalized chebyshev time (-1 .le. tc .le. 1)

    tc=2.d0*(dmod(temp,1.d0)+dt1)-1.d0

!       check to see whether chebyshev time has changed,
!       and compute new polynomial values if it has.
!       (the element pc(2) is the value of t1(tc) and hence
    !       contains the value of tc on the previous call.)
    
    if(tc.ne.pc(2)) then
       np=2
       nv=3
       pc(2)=tc
       twot=tc+tc
    endif
    !
    !       be sure that at least 'ncf' polynomials have been evaluated
    !       and are stored in the array 'pc'.
    !
    if(np.lt.ncf) then
       do i=np+1,ncf
          pc(i)=twot*pc(i-1)-pc(i-2)
       end do
       np=ncf
    endif
    !
    !       interpolate to get position for each component
    !
    do i=1,ncm
       pv(i,1)=0.d0
       do j=ncf,1,-1
          pv(i,1)=pv(i,1)+pc(j)*buf(j,i,l)
       end do
    end do
    if(ifl.le.1) return
    !
    !       if velocity interpolation is wanted, be sure enough
    !       derivative polynomials have been generated and stored.
    !
    vfac=(dna+dna)/t(2)
    vc(3)=twot+twot
    if(nv.lt.ncf) then
       do i=nv+1,ncf
          vc(i)=twot*vc(i-1)+pc(i-1)+pc(i-1)-vc(i-2)
       end do
       nv=ncf
    endif
    !
    !       interpolate to get velocity for each component
    !
    do i=1,ncm
       pv(i,2)=0.d0
       do j=ncf,2,-1
          pv(i,2)=pv(i,2)+vc(j)*buf(j,i,l)
       end do
       pv(i,2)=pv(i,2)*vfac
    end do
    !
    return

  end subroutine interp
  
end module jplephemeris_mod
